#   
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2021 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import annotations

import ast
import re
from typing import Optional, Sequence, TypeVar

import casadi as ca


Symbolic = TypeVar('Symbolic', ca.SX, ca.MX)


FUNCTIONS = {
    'sqrt': ca.sqrt,
    'exp': ca.exp,
    'log': ca.log,
    'log10': ca.log10,
    'sign': ca.sign,
    'abs': ca.fabs,
    'min': ca.fmin,
    'max': ca.fmax,
    'sin': ca.sin,
    'cos': ca.cos,
    'tan': ca.tan,
    'arcsin': ca.asin,
    'arccos': ca.acos,
    'arctan': ca.atan,
    'arctan2': ca.atan2,
    'sinh': ca.sinh,
    'cosh': ca.cosh,
    'tanh': ca.tanh,
    'arsinh': ca.asinh,
    'arcosh': ca.acosh,
    'artanh': ca.atanh
}


class Parser:
    """"""
    fx = ca.SX

    @classmethod
    def _get_op(cls, node):
        """

        :param node:
        :return:
        """
        op_type = type(node.op)

        return {
            ast.Add: ca.plus,
            ast.Sub: ca.minus,
            ast.Mult: ca.times,
            ast.Div: ca.rdivide,
            ast.Pow: ca.power,
            # ast.BitXor: ca.power,  # Doesn't make sense right now (see https://github.com/casadi/casadi/issues/586)
            ast.USub: neg
        }[op_type]

    @classmethod
    def _get_op_fun(cls, node):
        """

        :param node:
        :return:
        """
        return FUNCTIONS[node.func.id]

    @classmethod
    def _num_op(cls, node, subs, free):
        """

        :param node:
        :param subs:
        :param free:
        :return:
        """
        return node.n

    @classmethod
    def _bin_op(cls, node, subs, free):
        """

        :param node:
        :param subs:
        :param free:
        :return:
        """
        op = cls._get_op(node)
        left_node = cls.eval(node.left, subs, free)
        right_node = cls.eval(node.right, subs, free)
        return op(left_node, right_node)

    @classmethod
    def _unary_op(cls, node, subs, free):
        """

        :param node:
        :param subs:
        :param free:
        :return:
        """
        op = cls._get_op(node)
        return op(cls.eval(node.operand, subs, free))

    @classmethod
    def _subs_op(cls, node, subs, free):
        """

        :param node:
        :param subs:
        :return:
        """
        n_id = node.id
        if n_id in subs:
            return subs[n_id]
        else:
            if free is not None:
                value = cls.fx.sym(n_id)
                if isinstance(free, dict):
                    free[n_id] = value
                elif isinstance(free, (list, set, tuple)):
                    free.append(value)
                else:
                    msg = "Type {} for argument 'free' not supported".format(type(free))
                    raise TypeError(msg)
                subs[n_id] = value
                return value
            else:
                raise TypeError(node)

    @classmethod
    def _call_op(cls, node, subs, free):
        """

        :param node:
        :param subs:
        :param free:
        :return:
        """
        if node.func.id in FUNCTIONS:
            args = []
            for arg in node.args:
                args.append(cls.eval(arg, subs, free))
            fun = cls._get_op_fun(node)
            return fun(*args)
        else:
            # NOTE: Everything that is dependent on the time and not part of subs will be treated as a function
            # (like e.g. u(t)). We could also raise an error here.
            return cls.eval(node.func, subs, free)

    @classmethod
    def _const_op(cls, node, subs, free):
        """

        :param node:
        :param subs:
        :param free:
        :return:
        """
        return node.value

    @classmethod
    def eval(cls, node, subs, free):
        """

        :param node:
        :param subs:
        :param free:
        :return:
        """
        node_type = type(node)

        return {
            ast.Num: cls._num_op,
            ast.BinOp: cls._bin_op,
            ast.UnaryOp: cls._unary_op,
            ast.Name: cls._subs_op,
            ast.Call: cls._call_op,
            ast.Constant: cls._const_op
        }[node_type](node, subs, free)

    @classmethod
    def eval_expression(cls, expr, subs=None, free=None):
        """

        :param expr:
        :param subs:
        :param free:
        :return:
        """
        if subs is None:
            subs = {}

        expr_tree = ast.parse(expr, mode='eval').body
        return cls.eval(expr_tree, subs, free)


def neg(operand):
    """

    :param operand:
    :return:
    """
    return operand.__neg__()


def is_numeric_2(string):
    """

    :param string:
    :return:
    """
    try:
        value = float(string)
    except ValueError:
        return None
    else:
        return value


def parse_dynamic_equations(
        equations: Sequence[str],
        discrete: bool = False,
        replace_bitwise_xor: bool = True,
        use_sx: bool = True,
        **kwargs: Optional[Sequence[Symbolic]]
) -> dict[str, list[Symbolic]]:
    """

    :param equations:
    :param discrete:
    :param replace_bitwise_xor:
    :param use_sx:
    :param kwargs:
    :return:
    """
    # NOTE: This is a work in progress, that will be updated as more and more examples become available
    # TODO: Multi-mode parsing
    # TODO: Catch not working names, like e.g. lambda
    dt = kwargs.get('dt')
    t = kwargs.get('t')
    x = kwargs.get('x')
    y = kwargs.get('y')
    z = kwargs.get('z')
    u = kwargs.get('u')
    p = kwargs.get('p')

    to_substitute = {}
    if dt is not None:
        to_substitute.update({k.name(): k for k in dt})
    if t is not None:
        to_substitute.update({k.name(): k for k in t})
    if x is not None:
        to_substitute.update({k.name(): k for k in x})
    else:
        x = []
    if y is not None:
        to_substitute.update({k.name(): k for k in y})
    else:
        y = []
    if z is not None:
        to_substitute.update({k.name(): k for k in z})
    else:
        z = []
    if u is not None:
        to_substitute.update({k.name(): k for k in u})
    else:
        u = []
    if p is not None:
        to_substitute.update({k.name(): k for k in p})
    else:
        p = []

    description = {}
    labels = {}
    units = {}

    model = {
        'x': x,
        'y': y,
        'z': z,
        'u': u,
        'p': p,
        'ode': [],
        'alg': [],
        'quad': [],
        'meas': [],
        'const': {}
    }

    if use_sx:
        fx = ca.SX
        Parser.fx = ca.SX
    else:
        fx = ca.MX
        Parser.fx = ca.MX

    to_replace = {}
    odes = []
    algs = []
    quad = []
    meas = []

    ivp = '...'

    for k, equation in enumerate(equations):
        if '|' in equation:
            var, prop = equation.split('|')
            if ':' in prop:
                var = var.lstrip()
                key, val = prop.split(':')
                if key == 'description':
                    description[var] = val.lstrip()
                elif key == 'label':
                    labels[var] = val.lstrip()
                elif key == 'unit':
                    units[var] = val.lstrip()
            continue

        eq = equation.replace(' ', '')
        ivp = ivp.replace('...', eq)
        if ivp.endswith('...'):
            continue
        elif ivp.startswith('#') or '=' not in ivp:
            ivp = '...'
            continue
        if replace_bitwise_xor:
            ivp = ivp.replace('^', '**')
        else:
            raise NotImplementedError("Bitwise operations are not yet supported for parsing")

        lhs, rhs = ivp.split('=')

        if not discrete:
            # Check if equation is a quadrature function
            # TODO: Raise warning when 'sum' is used?
            if lhs == 'int':
                quad.append(rhs)
                continue

            # Check if equation is a differential equation
            # Check if left-hand side is of form dx(t)/dt
            match = re.search(r'd(.+)/dt', lhs)
            if match is None:
                # If there is not match, check if left-hand side is of form d/dt(x(t))
                match = re.search(r'd/dt\((.+)\)', lhs)
                if match is not None:
                    state = re.sub(r'\(.*?\)', '', match.groups()[0])
                    if state not in to_substitute:
                        value = fx.sym(state)
                        to_substitute[state] = value
                        model['x'].append(value)
                    # In case we have some nested expressions using d/dt(x(t))
                    to_replace[lhs] = f'({rhs})'
                    odes.append(rhs)
                else:
                    # No match, i.e. no differential equation
                    if lhs == '0':
                        # Algebraic equation 0 = g(x,z,u,t) (implicit form)
                        algs.append(rhs)
                    else:
                        value = is_numeric_2(rhs)
                        if value is None:
                            # Either algebraic variable or measurement
                            match = re.search(r'(.*)\(t\)', lhs)
                            if match is None:
                                # No algebraic variable
                                match = re.search(r'(.*)\(k\)', lhs)
                                if match is None:
                                    # No measurement, just an auxiliary variable that needs to be replaced
                                    # TODO: Process data even more, see DaeBuilder ddef (dependent parameters)
                                    to_replace[lhs] = f'({rhs})'
                                else:
                                    # Measurement y(k) = ...
                                    if match.groups()[0] not in to_substitute:
                                        value = fx.sym(match.groups()[0])
                                        to_substitute[match.groups()[0]] = value
                                        model['y'].append(value)
                                    meas.append(rhs)
                            else:
                                # Algebraic variable z(t) = ... (explicit form)
                                if match.groups()[0] not in to_substitute:
                                    value = fx.sym(match.groups()[0])
                                    to_substitute[match.groups()[0]] = value
                                    model['z'].append(value)
                                algs.append(rhs + '-' + lhs)
                        else:
                            # Constants
                            to_substitute[lhs] = value
                            model['const'][lhs] = value
            else:
                state = re.sub(r'\(.*?\)', '', match.groups()[0])
                if state not in to_substitute:
                    value = fx.sym(state)
                    to_substitute[state] = value
                    model['x'].append(value)
                # In case we have some nested expressions using d(x(t))/dt
                to_replace[lhs] = f'({rhs})'
                odes.append(rhs)
        else:
            if lhs == 'sum':
                # TODO: Raise warning when 'int' is used?
                quad.append(rhs)
                continue

            match = re.search(r'(.*)\(k\+1\)', lhs)
            if match is not None:
                # Difference equation
                if match.groups()[0] not in to_substitute:
                    value = fx.sym(match.groups()[0])
                    to_substitute[match.groups()[0]] = value
                    model['x'].append(value)
                # In case we have some nested expressions using x(k+1)
                to_replace[lhs] = f'({rhs})'
                odes.append(rhs)
            else:
                # No match, i.e. no difference equation
                if lhs == '0':
                    # Algebraic equation 0 = g(x(k),z(k),u(k),t(k)) (implicit form)
                    algs.append(rhs)
                else:
                    value = is_numeric_2(rhs)
                    if value is None:
                        # Check if measurement (algebraic variables are not possible here, since they would be
                        # indistinguishable from the measurements: y(k) vs z(k))
                        match = re.search(r'(.*)\(k\)', lhs)
                        if match is None:
                            # No measurement, just an auxiliary variable that needs to be replaced
                            # TODO: Process data even more, see DaeBuilder ddef (dependent parameters)
                            to_replace[lhs] = f'({rhs})'
                        else:
                            # Measurement y(k) = ...
                            if match.groups()[0] not in to_substitute:
                                value = fx.sym(match.groups()[0])
                                to_substitute[match.groups()[0]] = value
                                model['y'].append(value)
                            meas.append(rhs)

        ivp = '...'

    to_replace = dict((re.escape(k), v) for k, v in sorted(to_replace.items(), key=lambda item: item[0], reverse=True))
    pattern = re.compile('|'.join(to_replace.keys()))

    def replace_nested(m) -> str:
        """

        :param m:
        :return:
        """
        key = m.group(0)
        end_pos = m.regs[0][-1]
        if end_pos == len(m.string):
            # NOTE: This could be any character as long as it's inside the list of the next if-condition
            next_char = '+'
        else:
            next_char = m.string[end_pos]
        if next_char in ['+', '-', '*', '/', ')']:
            val = nested.get(re.escape(key), None)
            if val is not None:
                return pattern.sub(replace_nested, val)
        return key

    def check_right_hand_side():
        """

        :return:
        """
        # NOTE: The following pattern doesn't include algebraic variables and inputs that have an underscore in their
        #  name
        # pattern = re.compile(r'([a-zA-Z]+)\(([kt0-9+\-]{1,3})\)')
        # NOTE: r'([.+])\(... could match the whole equation or parts of it depending on where it finds the last
        #  brackets containing 'k' or 't'
        pattern = re.compile(r'([a-zA-Z0-9_]+)\(([kt0-9+\-]{1,3})\)')
        matches = pattern.findall(eq)
        for match in matches:
            if match[0] not in FUNCTIONS and match[0] not in to_substitute:
                if not discrete:
                    value = fx.sym(match[0])
                    to_substitute[match[0]] = value
                    if match[1] == 't':
                        model['z'].append(value)
                    else:
                        # TODO: If necessary, deal with different time points for u, i.e. k, k-1, k-2,...
                        model['u'].append(value)
                else:
                    # TODO: How to distinguish between algebraic variables and input?
                    #  Do 'difference algebraic equations' even make sense?
                    pass

    nested = {}
    for ode in odes:
        # TODO: Maybe set variable that is to be replaced to 'ode' or 'alg' and then grab the slice, which is already a
        #  SX variable, inside the Parser, so we don't have to parse this string again
        if pattern.pattern:
            nested.update(to_replace)
            eq = pattern.sub(replace_nested, ode)
        else:
            # No pattern that needs to be replaced
            eq = ode
        check_right_hand_side()
        model['ode'].append(Parser.eval_expression(eq, subs=to_substitute, free=model['p']))

    for alg in algs:
        if pattern.pattern:
            nested.update(to_replace)
            eq = pattern.sub(replace_nested, alg)
        else:
            eq = alg
        check_right_hand_side()
        model['alg'].append(Parser.eval_expression(eq, subs=to_substitute, free=model['p']))

    for q in quad:
        if pattern.pattern:
            nested.update(to_replace)
            eq = pattern.sub(replace_nested, q)
        else:
            eq = q
        check_right_hand_side()
        model['quad'].append(Parser.eval_expression(eq, subs=to_substitute, free=model['p']))

    for mea in meas:
        if pattern.pattern:
            nested.update(to_replace)
            eq = pattern.sub(replace_nested, mea)
        else:
            eq = mea
        check_right_hand_side()
        model['meas'].append(Parser.eval_expression(eq, subs=to_substitute, free=model['p']))

    model['description'] = {
        'x': [description[xk.name()] if xk.name() in description else '' for xk in x],
        'y': [description[yk.name()] if yk.name() in description else '' for yk in y],
        'z': [description[zk.name()] if zk.name() in description else '' for zk in z],
        'u': [description[uk.name()] if uk.name() in description else '' for uk in u],
        'p': [description[pk.name()] if pk.name() in description else '' for pk in p]
    }
    model['labels'] = {
        'x': [labels[xk.name()] if xk.name() in labels else '' for xk in x],
        'y': [labels[yk.name()] if yk.name() in labels else '' for yk in y],
        'z': [labels[zk.name()] if zk.name() in labels else '' for zk in z],
        'u': [labels[uk.name()] if uk.name() in labels else '' for uk in u],
        'p': [labels[pk.name()] if pk.name() in labels else '' for pk in p]
    }
    model['units'] = {
        'x': [units[xk.name()] if xk.name() in units else '' for xk in x],
        'y': [units[yk.name()] if yk.name() in units else '' for yk in y],
        'z': [units[zk.name()] if zk.name() in units else '' for zk in z],
        'u': [units[uk.name()] if uk.name() in units else '' for uk in u],
        'p': [units[pk.name()] if pk.name() in units else '' for pk in p]
    }

    return model



#   
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2021 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#

import casadi as ca

from ..modules.dynamic_model.dynamic_model import Model
from ..util.plotting import get_plot_backend


def cstr_schaffner_and_zeitz():
    """

    :return:
    """
    # Load plot backend
    plot_backend = get_plot_backend()

    # Initialize empty model
    model = Model(name='CSTR', plot_backend=plot_backend)

    # Define model equations
    equations = """
    # ODEs
    dx_1/dt = -a_1*x_1(t) + b_1*r
    dx_2/dt = -a_2*x_2(t) + b_2*r + g*u(k)

    # Measurement
    y(k) = x_2(t)

    # Algebraic equations
    r = (1 - x_1(t))*exp(-E/(1 + x_2(t)))
    """
    model.set_equations(equations=equations)

    return model


def cstr_seborg():
    """

    :return:
    """
    # Load plot backend
    plot_backend = get_plot_backend()

    # Initialize empty model
    model = Model(name='CSTR', plot_backend=plot_backend)

    # Define model equations
    equations = """
    # ODEs
    dC_A/dt = q_0/V*(C_Af - C_A(t)) - k_0*exp(-E/(R*T(t)))*C_A(t)
    dT/dt = q_0/V*(T_f - T(t)) - DeltaH_r*k_0/(rho*C_p)*exp(-E/(R*T(t)))*C_A(t) + UA/(V*rho*C_p)*(T_c(t) - T(t))
    dT_c/dt = (T_cr(k) - T_c(t))/tau

    # Measurements
    y(k) = C_A(t)

    # Constants
    R = 8.314

    # General
    C_A|description: concentration of substrate A
    C_A|label: concentration
    C_A|unit: mol/L
    T|description: tank temperature
    T|label: temperature
    T|unit: K
    T_c|description: coolant temperature
    T_c|label: temperature
    T_c|unit: K
    T_cr|description: coolant temperature reference
    T_cr|label: temperature
    T_cr|unit: K
    # ...parameters
    """
    model.set_equations(equations=equations)

    return model


def ecoli_D1210_conti(model='simple'):
    """
    Returns a hilo-mpc Model object.
    The model is an adaptation to a continuous batch process from the fedbach process proposed in

    Lee,  J.;  Ramirez,  W.  F.
    Mathematical  modelling  of  induced  foreign protein  production  by  recombinant  bacteria.
    Biotechnol.  Bioeng. 1992,39, 635-646.

    If model ='simple' the model does not have the two inducer states and the reactions rates are left as parameters
    If model= 'complex' the model has the two inducer states and the reaction rates are the same as in Lee et al.

    :return:
    """
    # Load plot backend
    plot_backend = get_plot_backend()

    if model == 'complex':
        model = Model(plot_backend=plot_backend, name='ecoli_D1210_complex')
        x = model.set_dynamical_states('X', 'S', 'P', 'I', 'ISF', 'IRF')
        u = model.set_inputs('DS', 'DI')
        p = model.set_parameters('Sf', 'If')
        model.set_measurements('mu', 'Rs', 'Rfp')

        # Unwrap states
        X = x[0]
        S = x[1]
        P = x[2]
        I = x[3]
        ISF = x[4]
        IRF = x[5]

        # Unwrap inputs
        u1 = u[0]
        u2 = u[1]

        # Unwrap parameters
        Sf = p[0]
        If = p[1]

        # Reaction rates
        phi = 0.407 * S / (0.108 + S + (S ** 2) / 14814.0)
        mu = phi * (ISF + (0.22 * IRF) / (0.22 + I))
        Rs = 2 * mu
        Rfp = phi * (0.0005 + I) / (0.022 + I)
        k1 = 0.09 * I / (0.034 + I)
        k2 = 0.09 * I / (0.034 + I)

        D = u1 + u2

        dX = mu * X - D * X
        dS = - Rs * X - D * S + u1 * Sf
        dP = Rfp * X - D * P
        dI = - D * I + u2 * If
        dISF = -k1 * ISF
        dIRF = k2 * (1 - IRF)

        model.set_dynamical_equations([dX, dS, dP, dI, dISF, dIRF])
        model.set_measurement_equations([mu, Rs, Rfp])

        return model
    elif model == 'simple':
        model = Model(plot_backend=plot_backend, name='ecoli_D1210_conti_simple')
        x = model.set_dynamical_states('X', 'S', 'P', 'I')
        u = model.set_inputs('DS', 'DI')
        p = model.set_parameters('Sf', 'If', 'mu', 'Rs', 'Rfp')

        # Unwrap states
        X = x[0]
        S = x[1]
        P = x[2]
        I = x[3]

        # Unwrap inputs
        DS = u[0]
        DI = u[1]

        # Unwrap parameters
        Sf = p[0]
        If = p[1]

        # Unknown reaction rates
        mu = p[2]
        Rs = p[3]
        Rfp = p[4]

        D_tot = DS + DI

        dX = mu * X - D_tot * X
        dS = - Rs * X - D_tot * S + DS * Sf
        dP = Rfp * X - D_tot * P
        dI = - D_tot * I + DI * If
        model.set_dynamical_equations([dX, dS, dP, dI])

        return model


def ecoli_D1210_fedbatch():
    """
    Returns a hilo-mpc Model object of fedbatch bioreactor containing genetically modified Escherichia Coli D1210
    The model is taken from

    Lee,  J.;  Ramirez,  W.  F.
    Mathematical  modelling  of  induced  foreign protein  production  by  recombinant  bacteria.
    Biotechnol.  Bioeng.1992,39, 635-646.

    :return:
    """
    # Load plot backend
    plot_backend = get_plot_backend()

    model = Model(plot_backend=plot_backend, name='ecoli_D1210_fedbatch_complex')
    x = model.set_dynamical_states('X', 'S', 'P', 'I', 'ISF', 'IRF', 'V')
    u = model.set_inputs('FeedS', 'FeedI')

    Sf = 100
    If = 4

    # Unwrap states
    X = x[0]
    S = x[1]
    P = x[2]
    I = x[3]
    ISF = x[4]
    IRF = x[5]
    V = x[6]

    # Unwrap inputs
    u1 = u[0]
    u2 = u[1]

    # Reaction rates
    phi = 0.407 * S / (0.108 + S + (S ** 2) / 14814.0)
    mu = phi * (ISF + (0.22 * IRF) / (0.22 + I))
    Rs = 2 * mu
    Rfp = phi * (0.0005 + I) / (0.022 + I)
    k1 = 0.09 * I / (0.034 + I)
    k2 = 0.09 * I / (0.034 + I)

    D = (u1 + u2) / V

    dX = mu * X - D * X
    dS = - Rs * X - D * S + u1 * Sf / V
    dP = Rfp * X - D * P
    dI = - D * I + u2 * If / V
    dISF = -k1 * ISF
    dIRF = k2 * (1 - IRF)
    dV = u1 + u2

    model.set_dynamical_equations([dX, dS, dP, dI, dISF, dIRF, dV])

    return model


def scerevisiae_SEY2102_fedbatch():
    """
    Returns a hilo-mpc Model object of fedbatch bioreactor containing Saccharomyces Cerevisiae SEY2102. The model is
    taken from
    Park, Seujeung, and W. Fred Ramirez.
    "Dynamics of foreign protein secretion from Saccharomyces cerevisiae."
    Biotechnology and bioengineering 33.3 (1989): 272-281.

    :return:
    """
    # Load plot backend
    plot_backend = get_plot_backend()

    model = Model(plot_backend=plot_backend, name='scerevisae_SEY2102_fedbatch')
    x = model.set_dynamical_states('bio', 's', 'pt', 'pm', 'v')
    F = model.set_inputs('F')
    # s0 = model.set_parameters('s0')
    s0 = 20  # g/l
    bio = x[0]
    s = x[1]
    pt = x[2]
    pm = x[3]
    V = x[4]

    mu = (21.87 * s) / ((s + 0.4) * (s + 62.5))
    fp = (s * ca.exp(-5 * s)) / (s + 0.1)
    phi = 4.75 * mu / (0.12 + mu)

    D = F / V

    dbio = mu * bio - D * bio
    ds = -7.3 * mu * bio - D * (s - s0)
    dpt = fp * bio - D * pt
    dpm = phi * (pt - pm) - D * pm
    dv = F

    model.set_dynamical_equations([dbio, ds, dpt, dpm, dv])
    model.set_measurements('bio', 's', 'pt', 'pm', 'v')
    model.set_measurement_equations(model.x)

    return model


__all__ = [
    'cstr_schaffner_and_zeitz',
    'cstr_seborg',
    'ecoli_D1210_conti',
    'ecoli_D1210_fedbatch',
    'scerevisiae_SEY2102_fedbatch'
]



#   
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2021 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import annotations

from abc import ABCMeta, abstractmethod
import copy
from itertools import product
from typing import Dict, List, Optional, Sequence, Tuple, TypeVar, Union
import warnings

import casadi as ca
import numpy as np
from scipy.special import factorial, gamma as gamma_fun

from ....util.machine_learning import Parameter, Hyperparameter
from ....util.util import is_list_like


IntArray = Union[int, Sequence[int]]
Numeric = Union[int, float]
NumArray = Union[Numeric, Sequence[Numeric], np.ndarray]
Bounds = Dict[str, Union[str, Tuple[Numeric, Numeric]]]
Cov = TypeVar('Cov', bound='Kernel')
Array = TypeVar('Array', ca.SX, ca.MX, np.ndarray)
Param = TypeVar('Param', bound=Parameter)


class Kernel(metaclass=ABCMeta):
    """
    Kernel base class

    This base class implements the basic methods that are used for all basic kernels, e.g. dot product kernels and
    radial basis function kernels like the squared exponential. It also implements the methods for the composition of
    kernels with the addition, multiplication and exponentiation operation.

    :Note: The Kernel classes return a CasADi function for the respective inputs when called. Each basic Kernel class
        implements its own scalar covariance function. This base class implements the call method where the respective
        scalar covariance functions are generalized to the form of covariance matrices which will be returned as CasADi
        SX function. The SX symbol was chosen in accordance with the CasADi docs which recommend that low-level
        expressions are faster computed for SX symbols.

    :param active_dims:
    :type active_dims: int, list of int, optional
    """
    def __init__(self, active_dims: Optional[IntArray] = None):
        """Constructor method"""
        self.active_dims = active_dims

    def __add__(self, other: Cov) -> 'Sum':
        """Addition method"""
        return Sum(self, other)

    def __mul__(self, other: Cov) -> 'Product':
        """Multiplication method"""
        return Product(self, other)

    def __pow__(self, power: Numeric, modulo: Optional[int] = None) -> 'Power':
        """Power method"""
        return Power(self, power)

    def __radd__(self, other: Cov) -> 'Sum':
        """Addition method (from the right side)"""
        return Sum(other, self)

    def __rmul__(self, other: Cov) -> 'Product':
        """Multiplication method (from the right side)"""
        return Product(other, self)

    def __str__(self) -> str:
        """String representation method"""
        message = f"{self.__class__.__name__} with: \n"
        for attribute, value in self.__dict__.items():
            message += f"\t {attribute}: {value} \n"
        return message

    def __call__(self, X: Array, X_bar: Optional[Array] = None) -> Array:
        """Calling method"""
        X_val = X
        X_bar_val = X_bar

        is_symbolic = False
        if isinstance(X, (ca.SX, ca.MX)):
            if X_bar is not None:
                if not isinstance(X_bar, (ca.SX, ca.MX)):
                    raise ValueError("X and X_bar need to have the same type")
            is_symbolic = True
        else:
            if X_bar is not None:
                if isinstance(X_bar, (ca.SX, ca.MX)):
                    raise ValueError("X and X_bar need to have the same type")

        X, X_bar, X_is_X_bar = _clean_input_matrices(X, X_bar)

        dimension_input_space = X.rows()
        if self.active_dims is None:
            active_dims = np.arange(dimension_input_space, dtype=np.int_)
        else:
            active_dims = np.atleast_1d(np.asarray(self.active_dims, dtype=np.int_))
        x = ca.SX.sym('x', dimension_input_space, 1)
        x_bar = ca.SX.sym('x_bar', dimension_input_space, 1)

        covariance_function = self.get_covariance_function(x, x_bar, active_dims)
        covariance_matrix = self.get_covariance_matrix(covariance_function, X, X_bar)

        if is_symbolic:
            hyperparameters = {parameter.name: parameter.SX for parameter in self.hyperparameters}
        else:
            hyperparameters = {parameter.name: parameter.log / 2. if 'variance' in parameter.name else parameter.log for
                               parameter in self.hyperparameters}

        if X_is_X_bar:
            K = covariance_matrix(X=X_val, X_bar=X_val, **hyperparameters)['covariance']
        else:
            K = covariance_matrix(X=X_val, X_bar=X_bar_val, **hyperparameters)['covariance']

        if is_symbolic:
            return K
        else:
            return K.full()

    @property
    def hyperparameters(self) -> List[Param]:
        """
        List of all hyperparameters in the kernel.

        This attribute can be easily used to access specific attributes of all hyperparameters in generator expressions
        or list comprehensions.
        """
        hyperparameters = []
        for attribute in self.__dict__.values():
            if isinstance(attribute, Parameter):
                hyperparameters.append(attribute)
        return hyperparameters

    @property
    def hyperparameter_names(self) -> List[str]:
        """

        :return:
        """
        names = []
        for attribute in self.__dict__.values():
            if isinstance(attribute, Parameter):
                names.append(attribute.name)
        return names

    @abstractmethod
    def get_covariance_function(self, x: ca.SX, x_bar: ca.SX, active_dims: np.ndarray) -> ca.Function:
        """

        :param x:
        :param x_bar:
        :param active_dims:
        :return:
        """
        pass

    def get_covariance_matrix(self, covariance_function: ca.Function, X: ca.SX, X_bar: ca.SX) -> ca.Function:
        """

        :param covariance_function:
        :param X:
        :param X_bar:
        :return:
        """
        observations_in_X = X.columns()
        observations_in_X_bar = X_bar.columns()
        K = ca.SX.sym('K', observations_in_X, observations_in_X_bar)
        hyperparameter_symbols = {hyperparameter.name: hyperparameter.SX for hyperparameter in self.hyperparameters}
        hyperparameter_names = [hyperparameter.name for hyperparameter in self.hyperparameters]

        # TODO: Could maybe be faster with map?
        for i, j in product(range(observations_in_X), range(observations_in_X_bar)):
            K[i, j] = covariance_function(x=X[:, i], x_bar=X_bar[:, j], **hyperparameter_symbols)['covariance']

        covariance_matrix = ca.Function(
            'K',
            [X, X_bar, *hyperparameter_symbols.values()],
            [K],
            ['X', 'X_bar', *hyperparameter_names],
            ['covariance']
        )

        return covariance_matrix

    def is_bounded(self) -> bool:  # pragma: no cover
        """

        :return:
        """
        for parameter in self.hyperparameters:
            bounds = parameter.log_bounds
            if bounds[0] != -ca.inf:
                return True
            if bounds[1] != ca.inf:
                return True
        return False

    def update(self, *args) -> None:
        """

        :param args:
        :return:
        """
        self.parent.update(*args)

    @staticmethod
    def constant(bias: Numeric = 1., bounds: Optional[Bounds] = None) -> Cov:
        """

        :param bias:
        :param bounds:
        :return:
        """
        return ConstantKernel(bias=bias, bounds=bounds)

    @staticmethod
    def squared_exponential(
            active_dims: Optional[IntArray] = None,
            signal_variance: Numeric = 1.,
            length_scales: NumArray = 1.,
            ard: bool = False,
            bounds: Optional[Bounds] = None
    ) -> Cov:
        """

        :param active_dims:
        :param signal_variance:
        :param length_scales:
        :param ard:
        :param bounds:
        :return:
        """
        return SquaredExponentialKernel(active_dims=active_dims, signal_variance=signal_variance,
                                        length_scales=length_scales, ard=ard, bounds=bounds)

    @staticmethod
    def exponential(
            active_dims: Optional[IntArray] = None,
            signal_variance: Numeric = 1.,
            length_scales: NumArray = 1.,
            ard: bool = False,
            bounds: Optional[Bounds] = None
    ) -> Cov:
        """

        :param active_dims:
        :param signal_variance:
        :param length_scales:
        :param ard:
        :param bounds:
        :return:
        """
        return ExponentialKernel(active_dims=active_dims, signal_variance=signal_variance, length_scales=length_scales,
                                 ard=ard, bounds=bounds)

    @staticmethod
    def matern_32(
            active_dims: Optional[IntArray] = None,
            signal_variance: Numeric = 1.,
            length_scales: NumArray = 1.,
            ard: bool = False,
            bounds: Optional[Bounds] = None
    ) -> Cov:
        """

        :param active_dims:
        :param signal_variance:
        :param length_scales:
        :param ard:
        :param bounds:
        :return:
        """
        return Matern32Kernel(active_dims=active_dims, signal_variance=signal_variance, length_scales=length_scales,
                              ard=ard, bounds=bounds)

    @staticmethod
    def matern_52(
            active_dims: Optional[IntArray] = None,
            signal_variance: Numeric = 1.,
            length_scales: NumArray = 1.,
            ard: bool = False,
            bounds: Optional[Bounds] = None
    ) -> Cov:
        """

        :param active_dims:
        :param signal_variance:
        :param length_scales:
        :param ard:
        :param bounds:
        :return:
        """
        return Matern52Kernel(active_dims=active_dims, signal_variance=signal_variance, length_scales=length_scales,
                              ard=ard, bounds=bounds)

    @staticmethod
    def rational_quadratic(
            active_dims: Optional[IntArray] = None,
            signal_variance: Numeric = 1.,
            length_scales: NumArray = 1.,
            alpha: Numeric = 1.,
            ard: bool = False,
            bounds: Optional[Bounds] = None
    ) -> Cov:
        """

        :param active_dims:
        :param signal_variance:
        :param length_scales:
        :param alpha:
        :param ard:
        :param bounds:
        :return:
        """
        return RationalQuadraticKernel(active_dims=active_dims, signal_variance=signal_variance,
                                       length_scales=length_scales, alpha=alpha, ard=ard, bounds=bounds)

    @staticmethod
    def piecewise_polynomial(
            degree: int,
            active_dims: Optional[IntArray] = None,
            signal_variance: Numeric = 1.,
            length_scales: NumArray = 1.,
            ard: bool = False,
            bounds: Optional[Bounds] = None
    ) -> Cov:
        """

        :param degree:
        :param active_dims:
        :param signal_variance:
        :param length_scales:
        :param ard:
        :param bounds:
        :return:
        """
        return PiecewisePolynomialKernel(degree, active_dims=active_dims, signal_variance=signal_variance,
                                         length_scales=length_scales, ard=ard, bounds=bounds)

    @staticmethod
    def polynomial(
            degree: int,
            active_dims: Optional[IntArray] = None,
            signal_variance: Numeric = 1.,
            offset: Numeric = 1.,
            bounds: Optional[Bounds] = None
    ) -> Cov:
        """

        :param degree:
        :param active_dims:
        :param signal_variance:
        :param offset:
        :param bounds:
        :return:
        """
        return PolynomialKernel(degree, active_dims=active_dims, signal_variance=signal_variance, offset=offset,
                                bounds=bounds)

    @staticmethod
    def linear(
            active_dims: Optional[IntArray] = None,
            signal_variance: Numeric = 1.,
            bounds: Optional[Bounds] = None
    ) -> Cov:
        """

        :param active_dims:
        :param signal_variance:
        :param bounds:
        :return:
        """
        return LinearKernel(active_dims=active_dims, signal_variance=signal_variance, bounds=bounds)

    @staticmethod
    def neural_network(
            active_dims: Optional[IntArray] = None,
            signal_variance: Numeric = 1.,
            weight_variance: Numeric = 1.,
            bounds: Optional[Bounds] = None
    ) -> Cov:
        """

        :param active_dims:
        :param signal_variance:
        :param weight_variance:
        :param bounds:
        :return:
        """
        return NeuralNetworkKernel(active_dims=active_dims, signal_variance=signal_variance,
                                   weight_variance=weight_variance, bounds=bounds)

    @staticmethod
    def periodic(
            active_dims: Optional[IntArray] = None,
            signal_variance: Numeric = 1.,
            length_scales: NumArray = 1.,
            period: Numeric = 1.,
            bounds: Optional[Bounds] = None
    ) -> Cov:
        """

        :param active_dims:
        :param signal_variance:
        :param length_scales:
        :param period:
        :param bounds:
        :return:
        """
        return PeriodicKernel(active_dims=active_dims, signal_variance=signal_variance, length_scales=length_scales,
                              period=period, bounds=bounds)


class ConstantKernel(Kernel):
    """
    Constant covariance function

    :param bias:
    :type bias:
    :param bounds:
    :type bounds:
    """
    acronym = "Const"

    def __init__(
            self,
            bias: Numeric = 1.,
            bounds: Optional[Bounds] = None
    ) -> None:
        """Constructor method"""
        super().__init__()

        hyper_kwargs = {}
        if bounds is not None:  # pragma: no cover
            bias_bounds = bounds.get('bias')
            if bias_bounds is not None:
                if bias_bounds == 'fixed':
                    hyper_kwargs['fixed'] = True
                else:
                    hyper_kwargs['bounds'] = bias_bounds
        self.bias = Hyperparameter(f'{self.acronym}.bias', value=bias, **hyper_kwargs)

    def get_covariance_function(self, x: ca.SX, x_bar: ca.SX, active_dims: np.ndarray) -> ca.Function:
        """
        Covariance function of constant as follows:
            bias^2

        :param x:
        :param x_bar:
        :param active_dims:
        :return:
        """
        log_bias = self.bias.SX

        covariance_function = ca.Function(
            'covariance',
            [x, x_bar, log_bias],
            [ca.exp(2 * log_bias)],
            ['x', 'x_bar', self.bias.name],
            ['covariance']
        )

        return covariance_function


class StationaryKernel(Kernel, metaclass=ABCMeta):
    """
    Base for stationary kernels

    In the isotropic form the characteristic length scale is the same for every input dimensions. By parameterization
    each input dimension can be given a length scales, thus enabling automatic relevance detection during training.

    :param active_dims:
    :type active_dims:
    :param length_scales:
    :type length_scales:
    :param ard:
    :type ard:
    :param bounds:
    :type bounds:
    """
    acronym = "Stat"

    def __init__(
            self,
            active_dims: Optional[IntArray] = None,
            length_scales: NumArray = 1.,
            ard: bool = False,
            bounds: Optional[Bounds] = None
    ) -> None:
        """Constructor method"""
        super().__init__(active_dims=active_dims)

        hyper_kwargs = {}
        if bounds is not None:  # pragma: no cover
            length_scale_bounds = bounds.get('length_scales')
            if length_scale_bounds is not None:
                if length_scale_bounds == 'fixed':
                    hyper_kwargs['fixed'] = True
                else:
                    hyper_kwargs['bounds'] = length_scale_bounds

        if ard and active_dims is None:
            raise ValueError("The key word 'ard' can only be set to True if the key word 'active_dims' was supplied")

        if active_dims is not None and is_list_like(length_scales):
            if len(active_dims) != len(length_scales):
                raise ValueError(f"Dimension mismatch between 'active_dims' ({len(active_dims)}) and the number of "
                                 f"length_scales ({len(length_scales)})")

        if not is_list_like(length_scales) and ard:
            length_scales = len(active_dims) * [length_scales]

        self.length_scales = Hyperparameter(f'{self.acronym}.length_scales', value=length_scales, **hyper_kwargs)

    def get_parameterized_length_scales(self, dimension_input_space: int, log_length_scales: ca.SX) -> ca.Function:
        """
        The parameterized notation allows for automatic relevance detection if the length scales are given as vector
        (the kernel is considered anisotropic).

        :param dimension_input_space:
        :param log_length_scales:
        :return:
        """
        if self.is_isotropic():
            M = ca.Function('M',
                            [log_length_scales],
                            [ca.exp(-2 * log_length_scales) * ca.SX.eye(dimension_input_space)])
        elif log_length_scales.numel() == dimension_input_space:
            M = ca.Function('M', [log_length_scales], [ca.diag(ca.exp(-2 * log_length_scales))])
        else:
            raise ValueError("Length scales vector dimension does not equal input space dimension.")
        return M

    def is_isotropic(self) -> bool:
        """

        :return:
        """
        return self.length_scales.is_scalar()


class GammaExponentialKernel(StationaryKernel):
    """
    Gamma-exponential covariance function

    :param active_dims:
    :type active_dims:
    :param signal_variance:
    :type signal_variance:
    :param alpha:
    :type alpha:
    :param gamma:
    :type gamma:
    :param length_scales:
    :type length_scales:
    :param ard:
    :type ard:
    :param bounds:
    :type bounds:
    """
    acronym = "GE"

    def __init__(
            self,
            active_dims: Optional[IntArray] = None,
            signal_variance: Numeric = 1.,
            alpha: Optional[Numeric] = None,
            gamma: Numeric = 1.,
            length_scales: NumArray = 1.,
            ard: bool = False,
            bounds: Optional[Bounds] = None
    ) -> None:
        """Constructor method"""
        super().__init__(active_dims=active_dims, length_scales=length_scales, ard=ard, bounds=bounds)

        if bounds is not None:  # pragma: no cover
            variance_bounds = bounds.get('signal_variance')
            gamma_bounds = bounds.get('gamma')
            alpha_bounds = bounds.get('alpha')
        else:  # pragma: no cover
            variance_bounds, gamma_bounds, alpha_bounds = None, None, None

        hyper_kwargs = {}
        if variance_bounds is not None:  # pragma: no cover
            if variance_bounds == 'fixed':
                hyper_kwargs['fixed'] = True
            else:
                hyper_kwargs['bounds'] = variance_bounds
        self.signal_variance = Hyperparameter(f'{self.acronym}.signal_variance', value=signal_variance, **hyper_kwargs)

        hyper_kwargs = {}
        if gamma == 2.:
            raise ValueError("Value of the hyperparameter 'gamma' is set to 2. Use the squared exponential kernel "
                             "instead.")
        gamma /= 2 - gamma
        if gamma_bounds is not None:  # pragma: no cover
            if gamma_bounds == 'fixed':
                if gamma <= 0. or gamma > 2.:
                    raise ValueError("Value of the hyperparameter 'gamma' needs to be between 0 and 2")
                hyper_kwargs['fixed'] = True
            else:
                gamma_bounds = list(gamma_bounds)
                if gamma_bounds[0] < 0.:
                    warnings.warn("Value of hyperparameter 'gamma' needs to be bigger than 0. Switching lower bound to "
                                  "0...")
                    gamma_bounds[0] = 0.
                if gamma_bounds[1] > 2.:
                    warnings.warn("Value of hyperparameter 'gamma' needs to be smaller than 2. Switching upper bound to"
                                  " 2...")
                    gamma_bounds[1] = 2.
                gamma_bounds[0] /= 2 - gamma_bounds[0]
                gamma_bounds[1] /= 2 - gamma_bounds[1]
                hyper_kwargs['bounds'] = tuple(gamma_bounds)
        self.gamma = Hyperparameter(f'{self.acronym}.gamma', value=gamma, **hyper_kwargs)

        if alpha is not None:
            hyper_kwargs = {}
            if alpha_bounds is not None:  # pragma: no cover
                if alpha_bounds == 'fixed':
                    hyper_kwargs['fixed'] = True
                else:
                    hyper_kwargs['bounds'] = alpha_bounds
            self.alpha = Hyperparameter(f'{self.acronym}.alpha', value=alpha, **hyper_kwargs)
        else:
            self.alpha = 1.

    def get_covariance_function(self, x: ca.SX, x_bar: ca.SX, active_dims: np.ndarray) -> ca.Function:
        """
        Covariance function of gamma exponential as follows:

        .. math::

            \sigma^2 \cdot \exp(-1/2 \cdot ((x-x_{bar})^T M (x-x_{bar}))^{\gamma-1})

        where :math:`M` is the parameterized matrix of length scales

        :param x:
        :param x_bar:
        :param active_dims:
        :return:
        """
        log_std = self.signal_variance.SX
        log_length_scales = self.length_scales.SX

        log_gamma = self.gamma
        if isinstance(log_gamma, Parameter):
            gamma_arg_name = [log_gamma.name]
            log_gamma = log_gamma.SX
            gamma_arg = [log_gamma]
        else:
            log_gamma = ca.DM(log_gamma)
            log_gamma = ca.log(log_gamma / (2 - log_gamma))
            gamma_arg = []
            gamma_arg_name = []
        p = 2 / (1 - ca.exp(-log_gamma))

        log_alpha = self.alpha
        if isinstance(log_alpha, Parameter):
            alpha_arg_name = [log_alpha.name]
            log_alpha = log_alpha.SX
            alpha_arg = [log_alpha]
        else:
            log_alpha = ca.log(log_alpha)
            alpha_arg = []
            alpha_arg_name = []

        M = self.get_parameterized_length_scales(active_dims.size, log_length_scales)
        d2 = _mahalanobis_distance_squared(x[active_dims], x_bar[active_dims], M(log_length_scales))

        covariance_function = ca.Function(
            'covariance',
            [x, x_bar, log_std, log_length_scales] + gamma_arg + alpha_arg,
            [ca.exp(2 * log_std - ca.exp(log_alpha) * d2 ** (p / 2))],
            ['x', 'x_bar', self.signal_variance.name, self.length_scales.name] + gamma_arg_name + alpha_arg_name,
            ['covariance']
        )

        return covariance_function


class SquaredExponentialKernel(GammaExponentialKernel):
    """
    Squared exponential covariance function

    :param active_dims:
    :type active_dims:
    :param signal_variance:
    :type signal_variance:
    :param length_scales:
    :type length_scales:
    :param ard:
    :type ard:
    :param bounds:
    :type bounds:
    """
    acronym = "SE"

    def __init__(
            self,
            active_dims: Optional[IntArray] = None,
            signal_variance: Numeric = 1.,
            length_scales: NumArray = 1.,
            ard: bool = False,
            bounds: Optional[Bounds] = None
    ) -> None:
        """Constructor method"""
        super().__init__(active_dims=active_dims, signal_variance=signal_variance, length_scales=length_scales, ard=ard,
                         bounds=bounds)

        self.gamma = 2.
        self.alpha = .5


class MaternKernel(StationaryKernel):
    """
    Base for Matérn class of covariance functions

    :param p:
    :type p:
    :param active_dims:
    :type active_dims:
    :param signal_variance:
    :type signal_variance:
    :param length_scales:
    :type length_scales:
    :param ard:
    :type ard:
    :param bounds:
    :type bounds:
    """
    acronym = "Matern"

    def __init__(
            self,
            p: int,
            active_dims: Optional[IntArray] = None,
            signal_variance: Numeric = 1.,
            length_scales: NumArray = 1.,
            ard: bool = False,
            bounds: Optional[Bounds] = None
    ) -> None:
        """Constructor method"""
        super().__init__(active_dims=active_dims, length_scales=length_scales, ard=ard, bounds=bounds)

        if bounds is not None:  # pragma: no cover
            variance_bounds = bounds.get('signal_variance')
        else:  # pragma: no cover
            variance_bounds = None

        hyper_kwargs = {}
        if variance_bounds is not None:  # pragma: no cover
            if variance_bounds == 'fixed':
                hyper_kwargs['fixed'] = True
            else:
                hyper_kwargs['bounds'] = variance_bounds
        self.signal_variance = Hyperparameter(f'{self.acronym}.signal_variance', value=signal_variance, **hyper_kwargs)

        self._p = p

    def get_covariance_function(self, x: ca.SX, x_bar: ca.SX, active_dims: np.ndarray) -> ca.Function:
        """

        :param x:
        :param x_bar:
        :param active_dims:
        :return:
        """
        nu = self._p + .5

        log_std = self.signal_variance.SX
        log_length_scales = self.length_scales.SX

        M = self.get_parameterized_length_scales(active_dims.size, log_length_scales)
        d2 = _mahalanobis_distance_squared(x[active_dims], x_bar[active_dims], M(log_length_scales))

        if self._p > 1:
            gamma_p_plus_1 = gamma_fun(self._p + 1)
            gamma_2p_plus_1 = gamma_fun(2 * self._p + 1)
            poly1d = [gamma_p_plus_1 / gamma_2p_plus_1 * factorial(self._p + k) / (
                        factorial(k) * factorial(self._p - k)) * 2. ** (self._p - k) for k in range(self._p - 1)]
        else:
            poly1d = []
        if self._p == 0:
            poly1d.append(0.)
        elif self._p >= 1.:
            poly1d.append(1.)
        for k in range(len(poly1d) - 1, 0, -1):
            poly1d[k - 1] /= poly1d[k]

        d = ca.sqrt(2 * nu * d2)
        f = 1. + d * poly1d[0]
        for k in range(1, len(poly1d)):
            f = 1. + d * poly1d[k] * f

        covariance_function = ca.Function(
            'covariance',
            [x, x_bar, log_std, log_length_scales],
            [ca.exp(2 * log_std - d) * f],
            ['x', 'x_bar', self.signal_variance.name, self.length_scales.name],
            ['covariance']
        )

        return covariance_function


class ExponentialKernel(MaternKernel):
    """
    Exponential covariance function

    :param active_dims:
    :type active_dims:
    :param signal_variance:
    :type signal_variance:
    :param length_scales:
    :type length_scales:
    :param ard:
    :type ard:
    :param bounds:
    :type bounds:
    """
    acronym = "E"

    def __init__(
            self,
            active_dims: Optional[IntArray] = None,
            signal_variance: Numeric = 1.,
            length_scales: NumArray = 1.,
            ard: bool = False,
            bounds: Optional[Bounds] = None
    ) -> None:
        """Constructor method"""
        super().__init__(0, active_dims=active_dims, signal_variance=signal_variance, length_scales=length_scales,
                         ard=ard, bounds=bounds)


class Matern32Kernel(MaternKernel):
    """
    Matérn covariance function where :math:`{\\nu}` = 3/2

    :param active_dims:
    :type active_dims:
    :param signal_variance:
    :type signal_variance:
    :param length_scales:
    :type length_scales:
    :param ard:
    :type ard:
    :param bounds:
    :type bounds:
    """
    acronym = "M32"

    def __init__(
            self,
            active_dims: Optional[IntArray] = None,
            signal_variance: Numeric = 1.,
            length_scales: NumArray = 1.,
            ard: bool = False,
            bounds: Optional[Bounds] = None
    ) -> None:
        """Constructor method"""
        super().__init__(1, active_dims=active_dims, signal_variance=signal_variance, length_scales=length_scales,
                         ard=ard, bounds=bounds)


class Matern52Kernel(MaternKernel):
    """
    Matérn covariance function where :math:`{\\nu}` = 5/2

    :param active_dims:
    :type active_dims:
    :param signal_variance:
    :type signal_variance:
    :param length_scales:
    :type length_scales:
    :param ard:
    :type ard:
    :param bounds:
    :type bounds:
    """
    acronym = "M52"

    def __init__(
            self,
            active_dims: Optional[IntArray] = None,
            signal_variance: Numeric = 1.,
            length_scales: NumArray = 1.,
            ard: bool = False,
            bounds: Optional[Bounds] = None
    ) -> None:
        """Constructor method"""
        super().__init__(2, active_dims=active_dims, signal_variance=signal_variance, length_scales=length_scales,
                         ard=ard, bounds=bounds)


class RationalQuadraticKernel(StationaryKernel):
    """
    Rational quadratic covariance function

    :param active_dims:
    :type active_dims:
    :param signal_variance:
    :type signal_variance:
    :param length_scales:
    :type length_scales:
    :param alpha:
    :type alpha:
    :param ard:
    :type ard:
    :param bounds:
    :type bounds:
    """
    acronym = "RQ"

    def __init__(
            self,
            active_dims: Optional[IntArray] = None,
            signal_variance: Numeric = 1.,
            length_scales: NumArray = 1.,
            alpha: Numeric = 1.,
            ard: bool = False,
            bounds: Optional[Bounds] = None
    ) -> None:
        """Constructor method"""
        super().__init__(active_dims=active_dims, length_scales=length_scales, ard=ard, bounds=bounds)

        if bounds is not None:  # pragma: no cover
            variance_bounds = bounds.get('signal_variance')
            alpha_bounds = bounds.get('alpha')
        else:  # pragma: no cover
            variance_bounds, alpha_bounds = None, None

        hyper_kwargs = {}
        if variance_bounds is not None:  # pragma: no cover
            if variance_bounds == 'fixed':
                hyper_kwargs['fixed'] = True
            else:
                hyper_kwargs['bounds'] = variance_bounds
        self.signal_variance = Hyperparameter(f'{self.acronym}.signal_variance', value=signal_variance, **hyper_kwargs)

        hyper_kwargs = {}
        if alpha_bounds is not None:  # pragma: no cover
            if alpha_bounds == 'fixed':
                hyper_kwargs['fixed'] = True
            else:
                hyper_kwargs['bounds'] = alpha_bounds
        self.alpha = Hyperparameter(f'{self.acronym}.alpha', value=alpha, **hyper_kwargs)

    def get_covariance_function(self, x: ca.SX, x_bar: ca.SX, active_dims: np.ndarray) -> ca.Function:
        """
        Covariance function of rational quadratic as follows:

        .. math::

            \sigma \cdot [1+1/(2\cdot\\alpha)\cdot(x-x_{bar})^T M (x-x_{bar})]^{-\\alpha}

        where :math:`M` is the parameterized matrix of length scales

        :param x:
        :param x_bar:
        :param active_dims:
        :return:
        """
        log_std = self.signal_variance.SX
        log_length_scales = self.length_scales.SX
        log_alpha = self.alpha.SX

        M = self.get_parameterized_length_scales(active_dims.size, log_length_scales)
        d2 = _mahalanobis_distance_squared(x[active_dims], x_bar[active_dims], M(log_length_scales))

        covariance_function = ca.Function(
            'covariance',
            [x, x_bar, log_std, log_length_scales, log_alpha],
            [ca.exp(2 * log_std) * (1 + .5 * d2 / ca.exp(log_alpha)) ** (-ca.exp(log_alpha))],
            # [ca.exp(2 * log_std - ca.exp(log_alpha) * ca.log(1 + .5 * d2 / ca.exp(log_alpha)))],
            ['x', 'x_bar', self.signal_variance.name, self.length_scales.name, self.alpha.name],
            ['covariance']
        )

        return covariance_function


class PiecewisePolynomialKernel(StationaryKernel):
    """
    Piecewise polynomial covariance function with compact support

    :param degree:
    :type degree:
    :param active_dims:
    :type active_dims:
    :param signal_variance:
    :type signal_variance:
    :param length_scales:
    :type length_scales:
    :param ard:
    :type ard:
    :param bounds:
    :type bounds:
    """
    acronym = "PP"

    def __init__(
            self,
            degree: int,
            active_dims: Optional[IntArray] = None,
            signal_variance: Numeric = 1.,
            length_scales: NumArray = 1.,
            ard: bool = False,
            bounds: Optional[Bounds] = None
    ) -> None:
        """Constructor method"""
        if degree not in [0, 1, 2, 3]:
            raise ValueError("The parameter 'degree' has to be one of the following integers: 0, 1, 2, 3")

        super().__init__(active_dims=active_dims, length_scales=length_scales, ard=ard, bounds=bounds)

        if bounds is not None:  # pragma: no cover
            variance_bounds = bounds.get('signal_variance')
        else:  # pragma: no cover
            variance_bounds = None

        hyper_kwargs = {}
        if variance_bounds is not None:  # pragma: no cover
            if variance_bounds == 'fixed':
                hyper_kwargs['fixed'] = True
            else:
                hyper_kwargs['bounds'] = variance_bounds
        self.signal_variance = Hyperparameter(f'{self.acronym}.signal_variance', value=signal_variance, **hyper_kwargs)

        self._q = degree

    @property
    def degree(self) -> int:
        """

        :return:
        """
        return self._q

    @degree.setter
    def degree(self, value: int) -> None:
        if value not in [0, 1, 2, 3]:
            raise ValueError("The property 'degree' has to be one of the following integers: 0, 1, 2, 3")
        self._q = value

    def get_covariance_function(self, x: ca.SX, x_bar: ca.SX, active_dims: np.ndarray) -> ca.Function:
        """

        :param x:
        :param x_bar:
        :param active_dims:
        :return:
        """
        q = self._q
        j = ca.floor(active_dims.size / 2) + q + 1

        log_std = self.signal_variance.SX
        log_length_scales = self.length_scales.SX

        M = self.get_parameterized_length_scales(active_dims.size, log_length_scales)
        d2 = _mahalanobis_distance_squared(x[active_dims], x_bar[active_dims], M(log_length_scales))
        d = ca.sqrt(d2)

        if q == 0:
            f = 1.
        elif q == 1:
            f = (j + 1) * d + 1
        elif q == 2:
            f = (j ** 2 + 4 * j + 3) / 3 * d2 + (j + 2) * d + 1
        elif q == 3:
            f = (j ** 3 + 9 * j ** 2 + 23 * j + 15) / 15 * d ** 3 + (6 * j ** 2 + 36 * j + 45) / 15 * d2 + (
                        j + 3) * d + 1
        else:
            raise RuntimeError("The parameter 'q' has to be one of the following integers: 0, 1, 2, 3")

        compact_support = (d < 1.) * ca.fmax(1. - d, 0.) ** (j + q)

        covariance_function = ca.Function(
            'covariance',
            [x, x_bar, log_std, log_length_scales],
            [ca.exp(2 * log_std) * compact_support * f],
            ['x', 'x_bar', self.signal_variance.name, self.length_scales.name],
            ['covariance']
        )

        return covariance_function


class DotProductKernel(Kernel, metaclass=ABCMeta):
    """
    Base for all dot product kernels

    :param active_dims:
    :type active_dims:
    :param signal_variance:
    :type signal_variance:
    :param offset:
    :type offset:
    :param bounds:
    :type bounds:
    """
    acronym = "Dot"

    def __init__(
            self,
            active_dims: Optional[IntArray] = None,
            signal_variance: Numeric = 1.,
            offset: Numeric = 1.,
            bounds: Optional[Bounds] = None
    ) -> None:
        """Constructor method"""
        super().__init__(active_dims=active_dims)

        if bounds is not None:  # pragma: no cover
            variance_bounds = bounds.get('signal_variance')
            offset_bounds = bounds.get('offset')
        else:  # pragma: no cover
            variance_bounds, offset_bounds = None, None

        hyper_kwargs = {}
        if variance_bounds is not None:  # pragma: no cover
            if variance_bounds == 'fixed':
                hyper_kwargs['fixed'] = True
            else:
                hyper_kwargs['bounds'] = variance_bounds
        self.signal_variance = Hyperparameter(f'{self.acronym}.signal_variance', value=signal_variance, **hyper_kwargs)

        hyper_kwargs = {}
        if offset_bounds is not None:  # pragma: no cover
            if offset_bounds == 'fixed':
                hyper_kwargs['fixed'] = True
            else:
                hyper_kwargs['bounds'] = offset_bounds
        self.offset = Hyperparameter(f'{self.acronym}.offset', value=offset, **hyper_kwargs)


class PolynomialKernel(DotProductKernel):
    """
    Polynomial covariance function

    :param degree:
    :type degree:
    :param active_dims:
    :type active_dims:
    :param signal_variance:
    :type signal_variance:
    :param offset:
    :type offset:
    :param bounds:
    :type bounds:
    """
    acronym = "Poly"

    def __init__(
            self,
            degree: int,
            active_dims: Optional[IntArray] = None,
            signal_variance: Numeric = 1.,
            offset: Numeric = 1.,
            bounds: Optional[Bounds] = None
    ) -> None:
        """Constructor method"""
        super().__init__(active_dims=active_dims, signal_variance=signal_variance, offset=offset, bounds=bounds)

        self._p = degree

    @property
    def degree(self) -> int:
        """

        :return:
        """
        return self._p

    @degree.setter
    def degree(self, value: int) -> None:
        self._p = value

    def get_covariance_function(self, x: ca.SX, x_bar: ca.SX, active_dims: np.ndarray) -> ca.Function:
        """
        Covariance function of polynomial as follows:
            (sigma_bias^2 + sigma_signal^2 * (x dot x_bar))^p

        :param x:
        :param x_bar:
        :param active_dims:
        :return:
        """
        p = self._p

        log_std = self.signal_variance.SX

        log_offset = self.offset
        if isinstance(log_offset, Parameter):
            offset_arg_name = [log_offset.name]
            log_offset = log_offset.SX
            offset_arg = [log_offset]
        else:
            log_offset = ca.log(log_offset)
            offset_arg = []
            offset_arg_name = []

        covariance_function = ca.Function(
            'covariance',
            [x, x_bar, log_std] + offset_arg,
            [ca.exp(2 * log_std) * (x[active_dims].T @ x_bar[active_dims] + ca.exp(log_offset)) ** p],
            ['x', 'x_bar', self.signal_variance.name] + offset_arg_name,
            ['covariance']
        )

        return covariance_function


class LinearKernel(PolynomialKernel):
    """
    Linear covariance function

    :param active_dims:
    :type active_dims:
    :param signal_variance:
    :type signal_variance:
    :param bounds:
    :type bounds:
    """
    acronym = "Lin"

    def __init__(
            self,
            active_dims: Optional[IntArray] = None,
            signal_variance: Numeric = 1.,
            bounds: Optional[Bounds] = None
    ) -> None:
        """Constructor method"""
        super().__init__(1, active_dims=active_dims, signal_variance=signal_variance, bounds=bounds)

        self.offset = 0.


class NeuralNetworkKernel(Kernel):
    """
    Neural network covariance function

    :param active_dims:
    :type active_dims:
    :param signal_variance:
    :type signal_variance:
    :param weight_variance:
    :type weight_variance:
    :param bounds:
    :type bounds:
    """
    acronym = "NN"

    def __init__(
            self,
            active_dims: Optional[IntArray] = None,
            signal_variance: Numeric = 1.,
            weight_variance: Numeric = 1.,
            bounds: Optional[Bounds] = None
    ) -> None:
        """Constructor method"""
        super().__init__(active_dims=active_dims)

        if bounds is not None:  # pragma: no cover
            signal_variance_bounds = bounds.get('signal_variance')
            weight_variance_bounds = bounds.get('weight_variance')
        else:  # pragma: no cover
            signal_variance_bounds, weight_variance_bounds = None, None

        hyper_kwargs = {}
        if signal_variance_bounds is not None:  # pragma: no cover
            if signal_variance_bounds == 'fixed':
                hyper_kwargs['fixed'] = True
            else:
                hyper_kwargs['bounds'] = signal_variance_bounds
        self.signal_variance = Hyperparameter(f'{self.acronym}.signal_variance', value=signal_variance, **hyper_kwargs)

        hyper_kwargs = {}
        if weight_variance_bounds is not None:  # pragma: no cover
            if weight_variance_bounds == 'fixed':
                hyper_kwargs['fixed'] = True
            else:
                hyper_kwargs['bounds'] = weight_variance_bounds
        self.weight_variance = Hyperparameter(f'{self.acronym}.weight_variance', value=weight_variance, **hyper_kwargs)

    def get_covariance_function(self, x: ca.SX, x_bar: ca.SX, active_dims: np.ndarray) -> ca.Function:
        """

        :param x:
        :param x_bar:
        :param active_dims:
        :return:
        """
        log_std = self.signal_variance.SX
        log_weight = self.weight_variance.SX

        num = 1. + x[active_dims].T @ x_bar[active_dims]
        den1 = ca.sqrt(ca.exp(2 * log_weight) + 1. + x[active_dims].T @ x[active_dims])
        den2 = ca.sqrt(ca.exp(2 * log_weight) + 1. + x_bar[active_dims].T @ x_bar[active_dims])

        covariance_function = ca.Function(
            'covariance',
            [x, x_bar, log_std, log_weight],
            [ca.exp(2 * log_std) * ca.asin(num / (den1 * den2))],
            ['x', 'x_bar', self.signal_variance.name, self.weight_variance.name],
            ['covariance']
        )

        return covariance_function


class PeriodicKernel(Kernel):
    """
    Periodic covariance function

    :param active_dims:
    :type active_dims:
    :param signal_variance:
    :type signal_variance:
    :param length_scales:
    :type length_scales:
    :param period:
    :type period:
    :param bounds:
    :type bounds:
    """
    acronym = "Periodic"

    def __init__(
            self,
            active_dims: Optional[IntArray] = None,
            signal_variance: Numeric = 1.,
            length_scales: Numeric = 1.,
            period: Numeric = 1,
            bounds: Optional[Bounds] = None
    ) -> None:
        """Constructor method"""
        super().__init__(active_dims=active_dims)

        if bounds is not None:  # pragma: no cover
            signal_variance_bounds = bounds.get('signal_variance')
            length_scales_bounds = bounds.get('length_scales')
            period_bounds = bounds.get('period')
        else:  # pragma: no cover
            signal_variance_bounds, length_scales_bounds, period_bounds = None, None, None

        hyper_kwargs = {}
        if signal_variance_bounds is not None:  # pragma: no cover
            if signal_variance_bounds == 'fixed':
                hyper_kwargs['fixed'] = True
            else:
                hyper_kwargs['bounds'] = signal_variance_bounds
        self.signal_variance = Hyperparameter(f'{self.acronym}.signal_variance', value=signal_variance, **hyper_kwargs)

        hyper_kwargs = {}
        if length_scales_bounds is not None:  # pragma: no cover
            if length_scales_bounds == 'fixed':
                hyper_kwargs['fixed'] = True
            else:
                hyper_kwargs['bounds'] = length_scales_bounds
        self.length_scales = Hyperparameter(f'{self.acronym}.length_scales', value=length_scales, **hyper_kwargs)

        hyper_kwargs = {}
        if period_bounds is not None:  # pragma: no cover
            if period_bounds == 'fixed':
                hyper_kwargs['fixed'] = True
            else:
                hyper_kwargs['bounds'] = period_bounds
        self.period = Hyperparameter(f'{self.acronym}.period', value=period, **hyper_kwargs)

    def get_covariance_function(self, x: ca.SX, x_bar: ca.SX, active_dims: np.ndarray) -> ca.Function:
        """
        Covariance function of exponential sine as follows:

        .. math::

            \sigma^2 \cdot \exp\\left(-1/2 \cdot \\frac{\sin(x-x_{bar})^T}{p} M \\frac{\sin(x-x_{bar})}{p}\\right)

        where :math:`M` is the parameterized matrix of length scales

        :param x:
        :param x_bar:
        :param active_dims:
        :return:
        """
        log_std = self.signal_variance.SX
        log_length_scales = self.length_scales.SX
        log_period = self.period.SX

        arg = ca.sin(ca.pi * (x[active_dims] - x_bar[active_dims]) / ca.exp(log_period)) / ca.exp(log_length_scales)

        covariance_function = ca.Function(
            'covariance',
            [x, x_bar, log_std, log_length_scales, log_period],
            [ca.exp(2 * log_std - 2 * arg ** 2)],
            ['x', 'x_bar', self.signal_variance.name, self.length_scales.name, self.period.name],
            ['covariance']
        )

        return covariance_function


class KernelOperator(Kernel, metaclass=ABCMeta):
    """
    Kernel operator base class

    There are three basic operations to composite kernels which preserve their properties as valid covariance matrices.
    These are addition, multiplication and exponentiation.

    :param kernel_1:
    :param kernel_2:
    """
    def __init__(self, kernel_1: Cov, kernel_2: Optional[Cov] = None) -> None:
        """Constructor method"""
        super().__init__()

        self.kernel_1 = copy.deepcopy(kernel_1)
        if kernel_2 is not None:
            self.kernel_2 = copy.deepcopy(kernel_2)
        else:
            self.kernel_2 = kernel_2
        self.disambiguate_hyperparameter_names()

    @property
    def hyperparameters(self) -> List[Param]:
        """
        List of all hyperparameters in the kernel.

        This attribute can be easily used to access specific attributes of all hyperparameters in generator expressions
        or list comprehensions.
        """
        if self.kernel_2 is not None:
            return self.kernel_1.hyperparameters + self.kernel_2.hyperparameters
        else:
            return self.kernel_1.hyperparameters

    @property
    def hyperparameter_names(self) -> List[str]:
        """

        :return:
        """
        if self.kernel_2 is not None:
            return self.kernel_1.hyperparameter_names + self.kernel_2.hyperparameter_names
        else:
            return self.kernel_1.hyperparameter_names

    def disambiguate_hyperparameter_names(self) -> None:
        """

        :return:
        """
        if self.kernel_2 is not None:
            kernel_1_has_acronym = hasattr(self.kernel_1, 'acronym')
            kernel_2_has_acronym = hasattr(self.kernel_2, 'acronym')
            if kernel_1_has_acronym and kernel_2_has_acronym:
                if self.kernel_1.acronym == self.kernel_2.acronym:
                    for parameter in self.kernel_1.hyperparameters:
                        old_name = parameter.name
                        if '.' in old_name:
                            new_name = self.kernel_1.acronym + '_1.' + old_name.split('.')[1]
                        else:
                            new_name = self.kernel_1.acronym + '_1.' + old_name
                        parameter.name = new_name
                    for parameter in self.kernel_2.hyperparameters:
                        old_name = parameter.name
                        if '.' in old_name:
                            new_name = self.kernel_2.acronym + '_2.' + old_name.split('.')[1]
                        else:
                            new_name = self.kernel_2.acronym + '_2.' + old_name
                        parameter.name = new_name
            elif kernel_1_has_acronym:
                kernel_2_acronyms = dict.fromkeys([name.split('.')[0] for name in self.kernel_2.hyperparameter_names])
                has_kernel_1_acronym = [self.kernel_1.acronym in name for name in kernel_2_acronyms]
                ct = 1
                for k, val in enumerate(kernel_2_acronyms.keys()):
                    if has_kernel_1_acronym[k]:
                        ct += 1
                        kernel_2_acronyms[val] = ct
                for parameter in self.kernel_1.hyperparameters:
                    old_name = parameter.name
                    if '.' in old_name:
                        new_name = self.kernel_1.acronym + '_1.' + old_name.split('.')[1]
                    else:
                        new_name = self.kernel_1.acronym + '_1.' + old_name
                    parameter.name = new_name
                for parameter in self.kernel_2.hyperparameters:
                    args = parameter.name.split('.')
                    old_acronym = args[0]
                    val = kernel_2_acronyms.get(old_acronym)
                    if val is not None:
                        new_acronym = old_acronym.split('_')[0] + '_' + str(val)
                        parameter.name = new_acronym + '.' + args[1]
            elif kernel_2_has_acronym:
                kernel_1_acronyms = set([name.split('.')[0] for name in self.kernel_1.hyperparameter_names])
                has_kernel_2_acronym = [self.kernel_2.acronym in name for name in kernel_1_acronyms]
                new_index = has_kernel_2_acronym.count(True) + 1
                for parameter in self.kernel_2.hyperparameters:
                    old_name = parameter.name
                    if '.' in old_name:
                        new_name = self.kernel_2.acronym + '_' + str(new_index) + '.' + old_name.split('.')[1]
                    else:
                        new_name = self.kernel_2.acronym + '_' + str(new_index) + '.' + old_name
                    parameter.name = new_name
            else:
                kernel_1_acronyms = set([name.split('.')[0] for name in self.kernel_1.hyperparameter_names])
                kernel_1_acronyms = [name.split('_')[0] for name in kernel_1_acronyms]
                kernel_2_acronyms = dict.fromkeys([name.split('.')[0] for name in self.kernel_2.hyperparameter_names])
                kernel_2_counter = {}
                for k, val in enumerate(kernel_2_acronyms.keys()):
                    args = val.split('_')
                    acronym = args[0]
                    count = kernel_1_acronyms.count(acronym)
                    kernel_2_count = kernel_2_counter.get(acronym)
                    if kernel_2_count is not None:
                        count += kernel_2_count
                        kernel_2_counter[acronym] += 1
                    else:
                        kernel_2_counter[acronym] = 1
                    if count != 0:
                        kernel_2_acronyms[val] = count + 1
                for parameter in self.kernel_2.hyperparameters:
                    args = parameter.name.split('.')
                    old_acronym = args[0]
                    val = kernel_2_acronyms.get(old_acronym)
                    if val is not None:
                        new_acronym = old_acronym.split('_')[0] + '_' + str(val)
                        parameter.name = new_acronym + '.' + args[1]


class Scale(KernelOperator):
    """"""
    def __init__(self, kernel: Cov, scale: Numeric) -> None:
        super().__init__(kernel, kernel_2=None)

        self.scale = scale


class Sum(KernelOperator):
    """
    Sum operator for covariance functions

    :param kernel_1:
    :type kernel_1:
    :param kernel_2:
    :type kernel_2:
    """
    def get_covariance_function(self, x: ca.SX, x_bar: ca.SX, active_dims: np.ndarray) -> ca.Function:
        """

        :param x:
        :param x_bar:
        :param active_dims:
        :return:
        """
        K_1 = self.kernel_1(x, x_bar)
        K_2 = self.kernel_2(x, x_bar)

        hyperparameters = [parameter.SX for parameter in self.hyperparameters]
        hyperparameter_names = [parameter.name for parameter in self.hyperparameters]

        K_sum = ca.Function(
            'covariance',
            [x, x_bar, *hyperparameters],
            [K_1 + K_2],
            ['x', 'x_bar', *hyperparameter_names],
            ['covariance']
        )

        return K_sum


class Product(KernelOperator):
    """
    Product operator for covariance functions

    :param kernel_1:
    :type kernel_1:
    :param kernel_2:
    :type kernel_2:
    """
    def get_covariance_function(self, x: ca.SX, x_bar: ca.SX, active_dims: np.ndarray) -> ca.Function:
        """

        :param x:
        :param x_bar:
        :param active_dims:
        :return:
        """
        K_1 = self.kernel_1(x, x_bar)
        K_2 = self.kernel_2(x, x_bar)

        hyperparameters = [parameter.SX for parameter in self.hyperparameters]
        hyperparameter_names = [parameter.name for parameter in self.hyperparameters]

        K_prod = ca.Function(
            'covariance',
            [x, x_bar, *hyperparameters],
            [K_1 * K_2],
            ['x', 'x_bar', *hyperparameter_names],
            ['covariance']
        )

        return K_prod


class Power(KernelOperator):
    """
    Power operator for covariance functions

    :param kernel:
    :type kernel:
    :param power:
    :type power:
    """
    def __init__(self, kernel: Cov, power: Numeric) -> None:
        super().__init__(kernel, kernel_2=None)

        self.power = power

    def get_covariance_function(self, x: ca.SX, x_bar: ca.SX, active_dims: np.ndarray) -> ca.Function:
        """

        :param x:
        :param x_bar:
        :param active_dims:
        :return:
        """
        K = self.kernel_1(x)
        power = self.power

        hyperparameters = [parameter.SX for parameter in self.hyperparameters]
        hyperparameter_names = [parameter.name for parameter in self.hyperparameters]

        K_pow = ca.Function(
            'covariance',
            [x, x_bar, *hyperparameters],
            [K ** power],
            ['x', 'x_bar', *hyperparameter_names],
            ['covariance']
        )

        return K_pow


class Warp(KernelOperator):
    """"""


def _clean_input_matrices(X: Array, X_bar: Optional[Array] = None) -> (ca.SX, ca.SX, bool):
    """
    Arbitrary array-like input matrix/matrices are converted to CasADi SX symbols. In the case of a single input
    matrix the resulting SX symbols are disambiguated, meaning they have unique CasADi identifiers.

    :param X:
    :param X_bar:
    :return:
    """
    if not isinstance(X, (ca.SX, ca.MX, ca.DM)):
        X = np.atleast_2d(X)
        X = ca.SX.sym('X', *X.shape)

    X_is_X_bar = False
    if X_bar is None or X_bar is X:
        X_bar = ca.SX.sym('X_bar', *X.shape)
        X_is_X_bar = True

    if not isinstance(X_bar, (ca.SX, ca.MX, ca.DM)):
        X_bar = np.atleast_2d(X_bar)
        X_bar = ca.SX.sym('X_bar', *X_bar.shape)

    assert X.shape[0] == X_bar.shape[0], "X and X_bar do not have the same input space dimensions"

    return X, X_bar, X_is_X_bar


def _mahalanobis_distance_squared(x, y, S):
    """

    :param x:
    :param y:
    :param S:
    :return:
    """
    x_minus_y = x - y
    return x_minus_y.T @ S @ x_minus_y


__all__ = [
    'Kernel',
    'ConstantKernel',
    # 'GammaExponentialKernel',
    'SquaredExponentialKernel',
    'MaternKernel',
    'ExponentialKernel',
    'Matern32Kernel',
    'Matern52Kernel',
    'RationalQuadraticKernel',
    'PiecewisePolynomialKernel',
    'DotProductKernel',
    'PolynomialKernel',
    'LinearKernel',
    'NeuralNetworkKernel',
    'PeriodicKernel'
]



#   
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2021 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#

from typing import Callable, Optional
import warnings

import casadi as ca
import numpy as np
from scipy.stats import norm

from .base import _Estimator
from ..dynamic_model.dynamic_model import Model
from ...util.util import convert


class ParticleFilter(_Estimator):
    """
    Particle filter (PF) class for state estimation

    :param model:
    :param id: The identifier of the PF object. If no identifier is given, a random one will be generated.
    :param name: The name of the PF object. By default the PF object has no name.
    :param plot_backend: Plotting library that is used to visualize estimated data. At the moment only
        `Matplotlib <https://matplotlib.org/>`_ and `Bokeh <https://bokeh.org/>`_ are supported. By default no plotting
        library is selected, i.e. no plots can be generated.
    :param variant:
    :param roughening:
    :param prior_editing:
    :param kwargs:
    """
    def __init__(
            self,
            model: Model,
            id: Optional[str] = None,
            name: Optional[str] = None,
            plot_backend: Optional[str] = None,
            variant: Optional[str] = None,
            roughening: bool = False,
            prior_editing: bool = False,
            **kwargs
    ):
        """Constructor method"""
        if model.is_linear():
            warnings.warn("The supplied model is linear. For better efficiency use an observer targeted at the "
                          "estimation of linear systems.")

        super().__init__(model, id=id, name=name, plot_backend=plot_backend)

        self._variant = variant
        self._roughening = roughening
        self._prior_editing = prior_editing
        if self._roughening or self._prior_editing:
            K = kwargs.get('K')
            if K is None:
                K = .2
            self._roughening_tuning_param = K

        self._setup_normpdf()
        self._sample_size = 15
        self._pdf = lhsnorm
        self._transpose_pdf = None

    def _update_type(self) -> None:
        """

        :return:
        """
        self._type = 'particle filter'

    def _setup_normpdf(self):
        """

        :return:
        """
        x = ca.SX.sym('x')
        mu = ca.SX.sym('mu')
        sigma = ca.SX.sym('sigma')

        y = ca.exp(-.5 * ((x - mu) / sigma) ** 2) / (ca.sqrt(2 * ca.pi) * sigma)

        self._normpdf = ca.Function('normpdf', [x, mu, sigma], [y])

    def _propagate_particles(self, n_samples):
        """

        :param n_samples:
        :return:
        """
        n_x = self._model.n_x
        n_y = self._model.n_y
        n_u = self._model.n_u
        n_p = self._model.n_p

        X = ca.MX.sym('X', n_x, n_samples)
        u = ca.MX.sym('u', n_u)
        p = ca.MX.sym('p', n_p)
        up = ca.vertcat(u, p)

        self._solution.setup('X', X={
            'values_or_names': [X.name() + '_' + str(k) for k in range(X.numel())],
            'description': X.numel() * [''],
            'labels': X.numel() * [''],
            'units': X.numel() * [''],
            'shape': (X.numel(), 0),
            'data_format': ca.DM
        })

        sol = self._model(x0=X, p=up)
        X_prop = sol['xf']
        if n_y == 0:
            warnings.warn(f"The model has no measurement equations, I am assuming measurements of all states "
                          f"{self._model.dynamical_state_names} are available.")
            Y = X_prop
        else:
            Y = sol['yf']

        w = ca.MX.sym('w', n_x, n_samples)
        v = ca.MX.sym('v', n_y, n_samples)

        # NOTE: At the moment only additive noise is supported
        X_prop += w
        Y += v

        self._predict_function = ca.Function('propagation_step',
                                             [X, up, w, v],
                                             [X_prop, Y],
                                             ['X', 'p', 'w', 'v'],
                                             ['X_prop', 'Y'])

    def _evaluate_likelihood(self, n_samples):
        """"""
        n_y = self._model.n_y

        y = ca.SX.sym('y', n_y)
        Y = ca.SX.sym('Y', n_y, n_samples)
        R = ca.SX.sym('R', (n_y, n_y))

        q = self._normpdf(Y, y, ca.sqrt(R))  # Since R is usually a diagonal matrix, we can use ca.sqrt() here. We need
        # to change this, if we have non-diagonal matrices, i.e. covariance entries.
        q /= ca.sum2(q)

        self._update_function = ca.Function('likelihood',
                                            [y, Y, R],
                                            [q],
                                            ['y', 'Y', 'R'],
                                            ['q'])

    def _initial_sample(self) -> None:
        """

        :return:
        """
        P0 = ca.reshape(self._solution.get_by_id('P:f'), self._n_x, self._n_x)
        X = self._pdf(self._solution.get_by_id('x:0').full().flatten(), P0, self._sample_size)
        if self._transpose_pdf:
            X = X.T
        elif self._transpose_pdf is None:
            if X.shape != (2, self._sample_size):
                X = X.T
                if X.shape != (2, self._sample_size):
                    raise ValueError(f"Dimension mismatch. Expected dimension 2x{self._sample_size}, got "
                                     f"{X.shape[1]}x{X.shape[0]}.")
                self._transpose_pdf = True
            else:
                self._transpose_pdf = False
        X = convert(X, ca.DM)
        self._solution.set('X', X[:])

    @property
    def probability_density_function(self) -> Callable[[np.ndarray, np.ndarray, int], np.ndarray]:
        """

        :return:
        """
        return self._pdf

    @probability_density_function.setter
    def probability_density_function(self, pdf: Callable[[np.ndarray, np.ndarray, int], np.ndarray]) -> None:
        """

        :param pdf:
        :return:
        """
        if not callable(pdf):
            raise ValueError(f"Probability density function of the {self.type} needs to be callable.")
        annotations = pdf.__annotations__
        run_function = True
        if annotations:
            if len(annotations) == 4 and 'return' in annotations:
                types = [np.ndarray, np.ndarray, int, np.ndarray]
                args = ["mean", "covariance", "sample size"]
                for k, type_ in enumerate(annotations.values()):
                    if type_ is not types[k]:
                        if k == 0:
                            no = "1st"
                        elif k == 1:
                            no = "2nd"
                        else:
                            no = "3rd"
                        if k < 3:
                            raise TypeError(f"The {no} argument to the probability density function (pdf) needs to be "
                                            f"the '{args[k]}' with type {types[k].__name__}.")
                        else:
                            raise TypeError(f"The return value of the probability density function (pdf) needs to be a"
                                            f" 'random sample' with type {types[k].__name__}.")
                run_function = False
        if run_function:
            try:
                X = pdf(np.zeros(self._n_x), np.eye(self._n_x), self._sample_size)
                if X.shape != (self._n_x, self._sample_size):
                    if X.shape != (self._sample_size, self._n_x):
                        raise ValueError(f"Dimension mismatch. Expected dimension {self._n_x}x{self._sample_size}, got "
                                         f"{X.shape[0]}x{X.shape[1]}.")
                    self._transpose_pdf = True
                else:
                    self._transpose_pdf = False
            except Exception as err:
                raise RuntimeError(f"The following exception was raised\n"
                                   f"   {type(err).__name__}: '{err.args[0]}'.\nPlease make sure that the "
                                   f"supplied probability density function (pdf) has the following arguments\n"
                                   f"   mu - mean of the pdf (type: numpy.ndarray),\n"
                                   f"   sigma - covariance of the mean (type: numpy.ndarray),\n"
                                   f"   n - sample size (type: int),\n"
                                   f"and the following return value\n"
                                   f"   X - random sample (type: numpy.ndarray).")

        self._pdf = pdf

    pdf = probability_density_function

    @property
    def variant(self) -> str:
        """

        :return:
        """
        return self._variant

    @variant.setter
    def variant(self, variant):
        self._variant = variant

    @property
    def sample_size(self):
        """

        :return:
        """
        return self._sample_size

    @sample_size.setter
    def sample_size(self, sample_size):
        self._sample_size = sample_size

    n_samples = sample_size

    def setup(self, **kwargs) -> None:
        """

        :param kwargs:
        :return:
        """
        self._solution.setup(self._model.solution)

        n_s = kwargs.get('n_samples')
        if n_s is None:
            n_s = self._sample_size
        else:
            self._sample_size = n_s

        self._propagate_particles(n_s)
        self._evaluate_likelihood(n_s)

        n_x = self._model.n_x
        n_y = self._model.n_y
        n_u = self._model.n_u
        n_p = self._model.n_p
        n_P = n_x * n_x

        self._solution.setup('P', P={
            'values_or_names': ['P_' + str(k) for k in range(n_P)],
            'description': n_P * [''],
            'labels': n_P * [''],
            'units': n_P * [''],
            'shape': (n_P, 0),
            'data_format': ca.DM
        })

        X = ca.MX.sym('X', n_x, n_s)
        y = ca.MX.sym('y', n_y)
        u = ca.MX.sym('u', n_u)
        p = ca.MX.sym('p', n_p)
        up = ca.vertcat(u, p)
        w = ca.MX.sym('w', n_x, n_s)
        v = ca.MX.sym('v', n_y, n_s)
        Q = ca.MX.sym('Q', (n_x, n_x))
        R = ca.MX.sym('R', (n_y, n_y))

        prediction = self._predict_function(X=X, p=up, w=w, v=v)
        X_prop = prediction['X_prop']
        Y = prediction['Y']
        update = self._update_function(y=y, Y=Y, R=R)

        self._function = ca.Function('function',
                                     [X, y, up, w, v, R],
                                     [X_prop, Y, update['q']],
                                     ['X', 'y', 'p', 'w', 'v', 'R'],
                                     ['X_prop', 'Y', 'q'])

        self._n_x = n_x
        self._n_y = n_y
        # self._n_z = n_z
        self._n_u = n_u
        self._n_p = n_p
        # self._n_p_est = n_p_est

        self._process_noise_covariance = ca.DM.zeros(Q.shape)
        self._measurement_noise_covariance = ca.DM.zeros(R.shape)

    def estimate(self, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        :return:
        """
        # TODO: Adjust according to Model.simulate
        self._check_setup()

        kwargs['skip'] = ['x']

        args = self._process_inputs(**kwargs)
        tf = args.pop('t0')
        steps = args.pop('steps')
        if self._solution.get_by_id('X').is_empty():
            self._initial_sample()
        args['X'] = ca.reshape(self._solution.get_by_id('X:f'), self._n_x, self._sample_size)

        if steps > 1:
            raise NotImplementedError("Particle filter is not yet implemented with steps > 1.")
        else:
            R = self._measurement_noise_covariance
            w = self._pdf(np.zeros(self._n_x), self._process_noise_covariance.full(), self._sample_size)
            # v = self._pdf(np.zeros(self._n_y), self._measurement_noise_covariance.full(), self._sample_size)
            v = ca.sqrt(R) @ np.random.randn(self._n_y, self._sample_size)  # Only possible since R is usually a
            # diagonal matrix (see self._evaluate_likelihood)
            if self._transpose_pdf:
                w = w.T
                # v = v.T
            args['w'] = w
            args['v'] = v
            args['R'] = R

            result = self._function(**args)
            Y = result['Y']

            # Prior editing
            if self._prior_editing:
                mag = args['y'] - Y
                need_roughening = ca.fabs(mag) > 6 * ca.sqrt(R)
                n_r = int(ca.sum2(need_roughening))
                while n_r > 0:
                    # TODO: Check out prior editing in more detail. Should we only use the X that are indexed by
                    #  need_roughening for the calculation of dx? Do we roughen already roughened X, or do we just
                    #  replace the ones that were improved by the roughening and keep the other ones at their original
                    #  value?
                    dx = np.max(args['X'], axis=1) - np.min(args['X'], axis=1)
                    dx = self._pdf(np.zeros(self._n_x), self._roughening_tuning_param * np.diag(dx) *
                                   n_r ** (-1 / self._n_x), n_r)
                    if self._transpose_pdf:
                        dx = dx.T
                    mask = np.where(need_roughening == 1)[1]
                    args['X'][:, mask] += dx

                    result = self._function(**args)

                    Y = result['Y']
                    mag = args['y'] - Y
                    need_roughening = ca.fabs(mag) > 6 * ca.sqrt(R)
                    n_r = int(ca.sum2(need_roughening))

            X = result['X_prop']
            q = result['q']

            # Resample (Survival of the fittest)
            ind = np.random.choice(self._sample_size, size=self._sample_size, replace=True, p=q.full().flatten())
            X = X[:, ind]
            Y = Y[:, ind]

            # Roughening
            if self._roughening:
                dx = np.max(X, axis=1) - np.min(X, axis=1)
                dx = self._pdf(np.zeros(self._n_x), self._roughening_tuning_param * np.diag(dx) *
                               self._sample_size ** (-1 / self._n_x), self._sample_size)
                if self._transpose_pdf:
                    dx = dx.T
                X += dx

            x = ca.sum2(X) / self._sample_size
            y = ca.sum2(Y) / self._sample_size
            P = convert(np.cov(X), ca.DM)
        self._solution.update(t=tf, x=x, X=X[:], P=P[:], y=y)


def lhsnorm(mu, sigma, n):
    """

    :param mu:
    :param sigma:
    :param n:
    :return:
    """
    # TODO: Migrate to somewhere in util
    n_m = mu.size
    z = np.random.multivariate_normal(mu, sigma, size=n)
    x = np.zeros_like(z, dtype=z.dtype)

    idz = np.argsort(z, axis=0)
    for k in range(n_m):
        x[idz[:, k], k] = np.linspace(1, n, n)
    x -= np.random.rand(*x.shape)
    x /= n

    for k in range(n_m):
        x[:, k] = norm.ppf(x[:, k], loc=mu[k], scale=np.sqrt(sigma[k, k]))

    return x


__all__ = [
    'ParticleFilter'
]



#
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2025 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#

"""
I/O plugins for industrial connectivity.

Currently includes an asyncio-based OPC UA client wrapper.
"""

from .opcua_async import AsyncOPCUAClient, IOMapping, build_mapping_from_model
from .opcua_connector import OPCUAConnector
from .opcua_simple_control_loop import OPCUASimpleControlLoop

__all__ = [
    "AsyncOPCUAClient",
    "IOMapping",
    "build_mapping_from_model",
    "OPCUAConnector",
    "OPCUASimpleControlLoop",
]



#
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2021 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#

from abc import ABCMeta, abstractmethod
from typing import Optional

import casadi as ca

from ..base import Base, Vector, TimeSeries
from ..dynamic_model.dynamic_model import Model
from ...util.util import convert


class Estimator(metaclass=ABCMeta):
    """"""
    _type = ''

    def __init__(self, *args, **kwargs):
        """Constructor method"""
        super().__init__(*args, **kwargs)
        self._update_type()

    @abstractmethod
    def _update_type(self) -> None:
        """

        :return:
        """
        pass

    @property
    def type(self) -> str:
        """

        :return:
        """
        return self._type


class _Estimator(Estimator, Base, metaclass=ABCMeta):
    """
    Base class for all estimators

    :param model:
    :param id: The identifier of the estimator. If no identifier is given, a random one will be generated.
    :param name: The name of the model. By default, the model has no name.
    :param plot_backend: Plotting library that is used to visualize estimated data. At the moment only Matplotlib and
        Bokeh are supported. By default, no plotting library is selected, i.e. no plots can be generated.
    """
    def __init__(
            self,
            model: Model,
            id: Optional[str] = None,
            name: Optional[str] = None,
            plot_backend: Optional[str] = None
    ) -> None:
        """Constructor method"""
        super().__init__(id=id, name=name)
        if self._id is None:
            self._create_id()

        if not model.is_setup():
            raise RuntimeError(f"Model is not set up. Run Model.setup() before passing it to the {self.type}.")

        self._model = model.copy(setup=True)
        self._error_covariance = None
        self._process_noise_covariance = None
        self._measurement_noise_covariance = None

        self._solution = TimeSeries(plot_backend, parent=self)

        self._n_x = 0
        self._n_y = 0
        self._n_z = 0
        self._n_u = 0
        self._n_p = 0
        self._n_p_est = 0

    def _set_process_noise(self, var):
        """

        :param var:
        :return:
        """
        Q = convert(var, ca.DM)
        if not Q.is_square():
            Q = ca.diag(Q)
        if Q.shape != (self._n_x, self._n_x):
            raise ValueError(f"Dimension mismatch. Supplied dimension is {Q.shape[0]}x{Q.shape[1]}, but required "
                             f"dimension is {self._n_x}x{self._n_x}")
        self._process_noise_covariance = Q

    def _set_measurement_noise(self, var):
        """

        :param var:
        :return:
        """
        R = convert(var, ca.DM)
        if not R.is_square():
            R = ca.diag(R)
        if R.shape != (self._n_y, self._n_y):
            raise ValueError(f"Dimension mismatch. Supplied dimension is {R.shape[0]}x{R.shape[1]}, but required "
                             f"dimension is {self._n_y}x{self._n_y}")
        self._measurement_noise_covariance = R

    def _set_error_covariance(self, P0):
        """

        :param P0:
        :return:
        """
        if P0 is None:
            P0 = ca.DM.eye(self._n_x)
        else:
            P0 = convert(P0, ca.DM)
            if not P0.is_square():
                P0 = ca.diag(P0)
            if P0.shape != (self._n_x, self._n_x):
                raise ValueError(f"Dimension mismatch. Supplied dimension is {P0.shape[0]}x{P0.shape[1]}, but required"
                                 f" dimension is {self._n_x}x{self._n_x}")
        self._solution.set('P', P0[:])

    def _check_setup(self):
        """

        :return:
        """
        if self._function is None:
            # NOTE: Not sure if we want to throw an error here
            type_ = self.type.split()
            type_[0] = type_[0].capitalize()
            msg = f"{' '.join(type_)} is not set up. Run {self.__class__.__name__}.setup() before running simulations."
            raise RuntimeError(msg)

        if self.n_x > 0 and self._solution.get_by_id('x').is_empty():
            raise RuntimeError(f"No initial guess for the states found. Please set initial guess before running the "
                               f"{self.type}!")

        if self._n_z > 0 and self._solution.get_by_id('z').is_empty():
            raise RuntimeError(f"No initial guess for the algebraic variables found. Please set initial guess before "
                               f"running the {self.type}!")

    def _process_inputs(self, **kwargs):
        """

        :param kwargs:
        :return:
        """
        dt = self._solution.dt
        if dt is not None:
            tf = kwargs.get('tf')
            if tf is None:
                steps = kwargs.get('steps', 1)
                tf = steps * dt
                # NOTE: Don't know if this is necessary
                tf = ca.linspace(dt, tf, steps).T
            else:
                steps = int(tf / dt)

                if self._solution['t'].is_empty():
                    self._solution.add('t', 0.)
        else:
            raise NotImplementedError("Support for grids is not yet implemented in the Kalman filter.")

        y = kwargs.get('y')
        u = kwargs.get('u')
        p = kwargs.get('p')
        if y is not None:
            if not isinstance(y, Vector):
                y = convert(y, ca.DM)
            if u is not None:
                if not isinstance(u, Vector):
                    u = convert(u, ca.DM)
            if p is not None:
                if not isinstance(p, Vector):
                    p = convert(p, ca.DM)
            if steps is not None:
                if y.size2() != 1 and steps != y.size2():
                    raise IndexError(f"Dimension mismatch for variable y. Supplied dimension is {y.size2()}, but "
                                     f"required dimension is {steps}.")
                elif y.size2() == 1 and steps != 1:
                    if isinstance(y, Vector):
                        y.repmat(steps, axis=1)
                    else:
                        y = ca.repmat(y, 1, steps)
            else:
                steps = y.size2()
            if u is not None:
                if u.size2() != 1 and steps != u.size2():
                    raise IndexError(f"Dimension mismatch for variable u. Supplied dimension is {u.size2()}, but "
                                     f"required dimension is {steps}.")
                elif u.size2() == 1 and steps != 1:
                    if isinstance(u, Vector):
                        u.repmat(steps, axis=1)
                    else:
                        u = ca.repmat(u, 1, steps)
                self._solution.add('u', u)
            if p is not None:
                if p.size2() != 1 and steps != p.size2():
                    raise IndexError(f"Dimension mismatch for variable p. Supplied dimension is {p.size2()}, but "
                                     f"required dimension is {steps}.")
                elif p.size2() == 1 and steps != 1:
                    if isinstance(p, Vector):
                        p.repmat(steps, axis=1)
                    else:
                        p = ca.repmat(p, 1, steps)
                self._solution.add('p', p)
        else:
            raise RuntimeError("No measurement data supplied.")

        skip = kwargs.get('skip')
        args = self._solution.get_function_args(steps=steps, skip=skip)
        args['t0'] += tf
        args['steps'] = steps
        # TODO: Add to solution like u and p?
        args['y'] = y

        return args

    @property
    def error_covariance(self):
        """

        :return:
        """
        if self._solution.is_set_up() and not self._solution.get_by_id('P').is_empty():
            return ca.reshape(self._solution['Pf'], self._n_x, self._n_x)
        return None

    P = error_covariance

    @property
    def process_noise_covariance(self):
        """

        :return:
        """
        return self._process_noise_covariance

    @process_noise_covariance.setter
    def process_noise_covariance(self, cov):
        self._set_process_noise(cov)

    Q = process_noise_covariance

    @property
    def measurement_noise_covariance(self):
        """

        :return:
        """
        return self._measurement_noise_covariance

    @measurement_noise_covariance.setter
    def measurement_noise_covariance(self, cov):
        self._set_measurement_noise(cov)

    R = measurement_noise_covariance

    @property
    def n_x(self):
        """

        :return:
        """
        return self._n_x

    @property
    def n_y(self):
        """

        :return:
        """
        return self._n_y

    @property
    def n_z(self):
        """

        :return:
        """
        return self._n_z

    @property
    def n_u(self):
        """

        :return:
        """
        return self._n_u

    @property
    def n_p(self):
        """

        :return:
        """
        return self._n_p

    @property
    def n_p_est(self):
        """

        :return:
        """
        return self._n_p_est

    @property
    def solution(self):
        """

        :return:
        """
        return self._solution

    def set_initial_guess(self, x0, t0=0., z0=None, P0=None):
        """

        :param x0:
        :param t0:
        :param z0:
        :param P0:
        :return:
        """
        self._model.set_initial_conditions(x0, t0=t0, z0=z0)
        self._solution.set('t', self._model.solution.get_by_id('t:0'))
        if not self._model.x.is_empty():
            self._solution.set('x', self._model.solution.get_by_id('x:0'))
        if not self._model.z.is_empty():
            self._solution.set('z', self._model.solution.get_by_id('z:0'))
        if not self._model.y.is_empty():
            self._solution.set('y', self._model.solution.get_by_id('y:0'))

        self._set_error_covariance(P0)

    def set_initial_parameter_values(self, p):
        """

        :param p:
        :return:
        """
        self._model.set_initial_parameter_values(p)
        if not self._model.p.is_empty():
            self._solution.add('p', p)

    @abstractmethod
    def setup(self, **kwargs):
        """

        :param kwargs:
        :return:
        """
        pass



#   
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2021 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#

from multiprocessing import Process
import os
import sys


class _TensorBoardSupervisor:
    """https://stackoverflow.com/a/60021949"""
    def __init__(self, log_dir, browser):
        """Constructor method"""
        self._server = _TensorBoardServer(log_dir)
        self._server.start()
        print("Started TensorBoard server")
        if browser == 'chrome':
            self._browser = _ChromeProcess()
        elif browser == 'firefox':
            self._browser = _FirefoxProcess()
        else:
            raise NotImplementedError(f"Support for browser '{browser}' not yet implemented")
        print("Started selected browser")
        self._browser.start()

    def finalize(self):
        """

        :return:
        """
        if self._server.is_alive():
            print("Killing TensorBoard server")
            self._browser.terminate()
            self._server.terminate()
            self._server.join()


class _TensorBoardServer(Process):
    """https://stackoverflow.com/a/60021949"""
    def __init__(self, log_dir):
        """Constructor method"""
        super().__init__()
        self._os_name = os.name
        self._log_dir = str(log_dir)

    def run(self) -> None:
        """

        :return:
        """
        if self._os_name == 'nt':  # Windows
            os.system(f'{sys.executable} -m tensorboard.main --logdir "{self._log_dir}" 2> NUL')
        elif self._os_name == 'posix':  # Linux
            os.system(f'{sys.executable} -m tensorboard.main --logdir "{self._log_dir}" --host `hostname -I` >/dev/null'
                      f' 2>&1')
        else:
            raise NotImplementedError(f"Support for OS '{self._os_name}' not yet implemented")


class _ChromeProcess(Process):
    """https://stackoverflow.com/a/60021949"""
    def __init__(self):
        """Constructor method"""
        super().__init__()
        self._os_name = os.name
        self.daemon = True

    def run(self) -> None:
        """

        :return:
        """
        if self._os_name == 'nt':  # Windows
            os.system(f'start chrome http://localhost:6006/')
        elif self._os_name == 'posix':  # Linux
            os.system('google-chrome http://$(hostname):6006/')
        else:
            raise NotImplementedError(f"Support for OS '{self._os_name}' not yet implemented")


class _FirefoxProcess(Process):
    """"""
    def __init__(self):
        """Constructor method"""
        super().__init__()
        self._os_name = os.name
        self.daemon = True

    def run(self) -> None:
        """

        :return:
        """
        if self._os_name == 'nt':  # Windows
            os.system('start firefox http://localhost:6006/')
        elif self._os_name == 'posix':  # Linux
            os.system('firefox http://localhost:6006/')
        else:
            raise NotImplementedError(f"Support for OS '{self._os_name}' not yet implemented")


class _TensorBoardWrapper:
    """"""
    def __init__(self, *args, **kwargs):
        """Constructor method"""
        self._log_dir = kwargs.get('log_dir')
        self._supervisor = None

    def _process_comment(self, prefix, **kwargs):
        """

        :param kwargs:
        :return:
        """
        comment = prefix + '-' + '-'.join(f'{key}={val}' for key, val in kwargs.items())
        path = os.path.join(self._log_dir, comment)  # TODO: What would happen here, if self._log_dir was None
        if not os.path.exists(path):
            os.makedirs(path)
        run_no = len(next(os.walk(path))[1])
        path += f'/run_{run_no + 1}'

        return comment, path

    @property
    def summary_writer(self):
        """

        :return:
        """
        from torch.utils.tensorboard import SummaryWriter

        return SummaryWriter

    @property
    def callback(self):
        """

        :return:
        """
        from tensorflow.keras.callbacks import TensorBoard

        return TensorBoard

    def get_summary_writer(self, **kwargs):
        """

        :param kwargs:
        :return:
        """
        comment, path = self._process_comment('pt', **kwargs)
        if self._log_dir is not None:
            return self.summary_writer(log_dir=path)
        else:
            return self.summary_writer(comment=comment)

    def get_callback(self, **kwargs):
        """

        :param kwargs:
        :return:
        """
        _, path = self._process_comment('tf', **kwargs)
        if self._log_dir is not None:
            return self.callback(log_dir=path)

    def show(self, browser):
        """

        :param browser:
        :return:
        """
        self._supervisor = _TensorBoardSupervisor(self._log_dir, browser)

    def close(self):
        """

        :return:
        """
        if self._supervisor is not None:
            self._supervisor.finalize()


def get_wrapper(*args, **kwargs):
    """

    :param args:
    :param kwargs:
    :return:
    """
    return _TensorBoardWrapper(*args, **kwargs)


__all__ = [
    'get_wrapper'
]



#   
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2021 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#

# TODO: Typing hints
# NOTE: Adapted from pandas plotting functionality

from abc import ABCMeta, abstractmethod
from math import ceil
import warnings

import matplotlib.cm as cm
import matplotlib.colors
import numpy as np

from ...util.util import is_list_like, random_state


_LINE_KWARGS = {
    'line_alpha': 'alpha',
    'line_cap': 'solid_capstyle',
    'line_color': 'color',
    'line_dash': 'linestyle',
    'line_join': 'solid_joinstyle',
    'line_width': 'linewidth'
}
_STEP_KWARGS = {
    'step_mode': 'where'
}
_SCATTER_KWARGS = {
    'marker': 'marker',
    'marker_color': 'c',
    'marker_size': 's'
}
_FILL_BETWEEN_KWARGS = {
    'line_color': 'edgecolor',
    'line_width': 'linewidth',
    'fill_color': 'color',
    'fill_alpha': 'alpha'
}


class BasePlot(metaclass=ABCMeta):
    """"""
    _layout_type = 'grid'
    _pop_attributes = [
        'tight_layout',
        'dpi',
        'background_fill_color'
    ]
    _attr_defaults = {
        'tight_layout': True,
        'dpi': None,
        'background_fill_color': None
    }

    def __init__(
            self,
            data,
            subplots=False,
            figsize=None,
            legend=False,
            legend_title=None,
            ax=None,
            title=None,
            xlabel=None,
            ylabel=None,
            colormap=None,
            layout=None,
            fill_between=None,
            interactive=False,
            anim=False,
            **kwargs
    ):
        """Constructor method"""
        self.data = data

        self.subplots = subplots

        self.figsize = figsize
        self.layout = layout
        self.fill_between = fill_between
        self.interactive = interactive
        self.anim = anim

        self.title = title
        self.xlabel = xlabel
        self.ylabel = ylabel

        self.legend = legend
        self.legend_title = legend_title
        self.legend_handles = []
        self.legend_labels = []

        for attr in self._pop_attributes:
            value = kwargs.pop(attr, self._attr_defaults.get(attr))
            setattr(self, attr, value)

        self.ax = ax

        if 'cmap' in kwargs and colormap:
            raise TypeError("Only specify one of 'cmap' and 'colormap'")
        elif 'cmap' in kwargs:
            self.colormap = kwargs.pop('cmap')
        else:
            self.colormap = colormap

        self.kwargs = kwargs

    def _add_legend_handle(self, handle, label):
        """

        :param handle:
        :param label:
        :return:
        """
        if label is not None:
            self.legend_handles.append(handle)
            self.legend_labels.append(label)

    def _decorate_subplots(self):
        """

        :return:
        """
        if is_list_like(self.xlabel):
            if len(self.xlabel) != self.n_plots:
                msg = f"The length of 'xlabel' must equal the number of columns if using 'xlabel' of type " \
                      f"'list'.\nlength of xlabel = {len(self.xlabel)}\nnumber of columns = {self.n_plots}"
                raise ValueError(msg)

        if is_list_like(self.ylabel):
            if len(self.ylabel) != self.n_plots:
                msg = f"The length of 'ylabel' must equal the number of columns if using 'ylabel' of type " \
                      f"'list'.\nlength of ylabel = {len(self.ylabel)}\nnumber of columns = {self.n_plots}"
                raise ValueError(msg)

        for k, ax in enumerate(self.axes):
            if ax.get_visible():
                if is_list_like(self.xlabel):
                    if self.xlabel[k] is not None:
                        ax.set_xlabel(self.xlabel[k])
                else:
                    if self.xlabel is not None:
                        ax.set_xlabel(self.xlabel)

                if is_list_like(self.ylabel):
                    if self.ylabel[k] is not None:
                        ax.set_ylabel(self.ylabel[k])
                else:
                    if self.ylabel is not None:
                        ax.set_ylabel(self.ylabel)

                if self.interactive:
                    _update_slider(self.fig, ax, self.sliders[k], self.data[k])

        if self.title:
            if self.subplots:
                if is_list_like(self.title):
                    if len(self.title) != self.n_plots:
                        msg = f"The length of 'title' must equal the number of columns if using 'title' of type " \
                              f"'list' and 'subplots=True'.\nlength of title = {len(self.title)}\n" \
                              f"number of columns = {self.n_plots}"
                        raise ValueError(msg)

                    for ax, title in zip(self.axes, self.title):
                        ax.set_title(title)
                else:
                    self.fig.suptitle(self.title)
            else:
                if is_list_like(self.title):
                    msg = "Using 'title' of type 'list' is not supported unless 'subplots=True' is passed."
                    raise ValueError(msg)
                self.axes[0].set_title(self.title)

    def _get_ax(self, k):
        """

        :param k:
        :return:
        """
        if self.subplots:
            ax = self.axes[k]
        else:
            ax = self.axes[0]

        ax.get_yaxis().set_visible(True)
        return ax

    @staticmethod
    def _get_ax_legend_handle(ax):
        """

        :param ax:
        :return:
        """
        legend = ax.get_legend()

        handle, _ = ax.get_legend_handles_labels()
        other_ax = getattr(ax, 'left_ax', None) or getattr(ax, 'right_ax', None)
        other_legend = None
        if other_ax is not None:
            other_legend = other_ax.get_legend()
        if legend is not None and other_legend is not None:
            legend = other_legend
            ax = other_ax
        return ax, legend, handle

    def _get_colors(self, num_colors=None, color_kwargs='color'):
        """

        :param num_colors:
        :param color_kwargs:
        :return:
        """
        if num_colors is None:
            num_colors = self.n_plots

        return _get_standard_colors(num_colors=num_colors, colormap=self.colormap, color=self.kwargs.get(color_kwargs))

    def _iter_data(self):
        """

        :return:
        """
        for sub, plots in self.data.items():
            index = plots['x']
            if 'line_color' in self.kwargs:
                colors = self._get_colors(num_colors=len(plots['y']), color_kwargs='line_color')
            elif 'marker_color' in self.kwargs:
                colors = self._get_colors(num_colors=len(plots['y']), color_kwargs='marker_color')
            else:
                colors = self._get_colors(num_colors=len(plots['y']))
            keys = list(plots['y'])
            for key, values in plots['y'].items():
                if isinstance(colors, dict):
                    color = colors[key]
                else:
                    color = colors[keys.index(key)]
                yield sub, index[key], key, values, color

    def _make_fill_between(self):
        """

        :return:
        """
        if not self.subplots:
            if is_list_like(self.fill_between):
                fill_between = self.fill_between[0]
            else:
                fill_between = self.fill_between
            if fill_between is not None:
                ax = self._get_ax(0)
                args = self._process_fill_between(fill_between)
                ax.fill_between(*args, **fill_between)
        else:
            if is_list_like(self.fill_between):
                if len(self.fill_between) != self.n_plots:
                    msg = f"The length of 'fill_between' must equal the number of columns if using 'fill_between' of " \
                          f"type 'list'.\nlength of fill_between = {len(self.fill_between)}\nnumber of columns = " \
                          f"{self.n_plots}"
                    raise ValueError(msg)
            for k in range(self.n_plots):
                if is_list_like(self.fill_between):
                    fill_between = self.fill_between[k]
                elif self.fill_between is not None:
                    fill_between = self.fill_between.copy()
                else:
                    fill_between = None
                if fill_between is not None:
                    ax = self._get_ax(k)
                    args = self._process_fill_between(fill_between)
                    ax.fill_between(*args, **fill_between)

    def _make_legend(self):
        """

        :return:
        """
        ax, legend, handle = self._get_ax_legend_handle(self.axes[0])

        handles = []
        labels = []
        title = ""

        if not self.subplots:
            if legend is not None:
                title = legend.get_title().get_text()
                handles.extend(handle)
                labels = [k.get_text() for k in legend.get_texts()]

            if self.legend:
                if self.legend == 'reverse':
                    self.legend_handles = reversed(self.legend_handles)
                    self.legend_labels = reversed(self.legend_labels)

                handles += self.legend_handles
                labels += self.legend_labels

                if self.legend_title is not None:
                    title = self.legend_title

            if len(handles) > 0:
                ax.legend(handles, labels, loc='best', title=title)

        elif self.subplots and self.legend:
            for k, ax in enumerate(self.axes):
                if ax.get_visible() and (self.legend or (is_list_like(self.legend) and self.legend[k])):
                    ax.legend(loc='best')

    @abstractmethod
    def _make_plot(self):
        """

        :return:
        """
        pass

    @classmethod
    def _plot(cls, ax, x, y, style=None, **kwargs):
        """

        :param ax:
        :param x:
        :param y:
        :param style:
        :param kwargs:
        :return:
        """
        if style is not None:
            args = (x, y, style)
        else:
            args = (x, y)
        return ax.plot(*args, **kwargs)

    @staticmethod
    def _process_fill_between(kwargs):
        """

        :param kwargs:
        :return:
        """
        x = kwargs.pop('x', None)
        if x is None:
            raise KeyError("No base for 'fill_between' option supplied")
        if x.ndim > 1:
            if 1 in x.shape:
                x = x.flatten()
            else:
                raise ValueError("Base argument for the 'fill_between' option needs to be a flat array")
        lb = kwargs.pop('lb', None)
        if lb is None:
            raise KeyError("No lower bound for 'fill_between' option supplied")
        if lb.ndim > 1:
            if 1 in lb.shape:
                lb = lb.flatten()
            else:
                raise ValueError("Lower bound argument for the 'fill_between' option needs to be a flat array")
        ub = kwargs.pop('ub', None)
        if ub is None:
            raise KeyError("No upper bound for 'fill_between' option supplied")
        if ub.ndim > 1:
            if 1 in ub.shape:
                ub = ub.flatten()
            else:
                raise ValueError("Upper bound argument for the 'fill_between' option needs to be a flat array")

        if x.size == lb.size == ub.size:
            kwargs['y2'] = lb
            for kw in _FILL_BETWEEN_KWARGS:
                if kw in kwargs:
                    kwargs[_FILL_BETWEEN_KWARGS[kw]] = kwargs.pop(kw)
            return x, ub
        else:
            raise ValueError(f"Dimension mismatch between base, lower and upper bound for the 'fill_between' option."
                             f"\nbase: {x.size}\nlower: {lb.size}\nupper: {ub.size}")

    def _setup_subplots(self):
        """

        :return:
        """
        fig_kwargs = {}
        subplot_kwargs = {}
        if self.figsize is not None:
            if isinstance(self.figsize, (list, tuple, set)) and len(self.figsize) <= 2:
                if len(self.figsize) == 1:
                    width = height = self.figsize[0]
                else:
                    width = self.figsize[0]
                    height = self.figsize[1]
                fig_kwargs['figsize'] = (width, height)

            if self.dpi is not None:
                fig_kwargs['dpi'] = self.dpi

        if self.background_fill_color is not None:
            subplot_kwargs['facecolor'] = self.background_fill_color
        fig_kwargs['tight_layout'] = self.tight_layout

        if self.subplots:
            fig_kwargs = {}
            if self.interactive:
                fig_kwargs['interactive'] = True
                fig_kwargs['x_data'] = [data['x'] for data in self.data.values()]
                fig_kwargs['tight_layout'] = False
            fig, axes, sliders = _subplots(n_axes=self.n_plots, layout=self.layout, layout_type=self._layout_type,
                                           **fig_kwargs)
        else:
            if self.ax is None:
                fig = self.plt.figure(**fig_kwargs)
                axes = fig.add_subplot(111, **subplot_kwargs)
            else:
                fig = self.ax.get_figure()
                if self.figsize is not None:
                    fig.set_size_inches(self.figsize)
                axes = self.ax
            sliders = []
            if self.interactive:
                sliders.append(_slider(axes, self.data[0]['x']))

        axes = _flatten(axes)

        self.fig = fig
        self.axes = axes
        self.sliders = sliders

    @property
    def n_plots(self):
        """

        :return:
        """
        return len(self.data)

    @property
    def plt(self):
        """

        :return:
        """
        import matplotlib.pyplot as plt

        return plt

    @property
    def result(self):
        """

        :return:
        """
        if self.subplots:
            if self.layout is not None:
                axes = self.axes.reshape(*self.layout)
            else:
                axes = self.axes
        else:
            axes = self.axes[0]

        if self.anim:
            from matplotlib.animation import FuncAnimation

            return self.plt, self.fig, axes, FuncAnimation
        else:
            return axes

    def draw(self):
        """

        :return:
        """
        if not self.anim:
            plt = self.plt
            plt.draw_if_interactive()
            plt.show()

    def generate(self):
        """

        :return:
        """
        self._setup_subplots()
        self._make_plot()
        self._make_fill_between()
        self._make_legend()
        self._decorate_subplots()


class MultiPlot(BasePlot):
    """"""
    def __init__(self, data, kind, **kwargs):
        """Constructor method"""
        super().__init__(data, **kwargs)

        self.kind = kind
        if len(self.kind) != self.n_plots:
            msg = f"The length of 'kind' must equal the number of columns if using 'kind' of type " \
                  f"'list'.\nlength of kind = {len(self.kind)}\nnumber of columns = {self.n_plots}"
            raise ValueError(msg)

        self._organize_kwargs()

    def _organize_kwargs(self):
        """

        :return:
        """
        supplied_kinds = set()
        for kind in self.kind:
            if is_list_like(kind):
                supplied_kinds.update(kind)
            else:
                supplied_kinds.add(kind)

        self._line_kwargs = {}
        self._step_kwargs = {}
        self._scatter_kwargs = {}

        for kw, val in self.kwargs.items():
            if kw in _LINE_KWARGS:
                if 'line' in supplied_kinds:
                    self._line_kwargs[_LINE_KWARGS[kw]] = val
                if 'dashed' in supplied_kinds:
                    self._line_kwargs[_LINE_KWARGS[kw]] = val
                if 'dotted' in supplied_kinds:
                    self._line_kwargs[_LINE_KWARGS[kw]] = val
                if 'dotdash' in supplied_kinds:
                    self._line_kwargs[_LINE_KWARGS[kw]] = val
                if 'dashdot' in supplied_kinds:
                    self._line_kwargs[_LINE_KWARGS[kw]] = val
                if 'step' in supplied_kinds:
                    self._step_kwargs[_LINE_KWARGS[kw]] = val
            if kw in _STEP_KWARGS:
                if 'step' in supplied_kinds:
                    self._step_kwargs[_STEP_KWARGS[kw]] = val
            if kw in _SCATTER_KWARGS:
                if 'scatter' in supplied_kinds:
                    self._scatter_kwargs[_SCATTER_KWARGS[kw]] = val
            if kw == 'x_range':
                self._line_kwargs[kw] = val
                self._step_kwargs[kw] = val
                self._scatter_kwargs[kw] = val

    def _get_kwargs(self, kind):
        """

        :param kind:
        :return:
        """
        if kind in ['line', 'dashed', 'dotted', 'dotdash', 'dashdot']:
            return self._line_kwargs.copy()
        elif kind == 'step':
            return self._step_kwargs.copy()
        elif kind == 'scatter':
            return self._scatter_kwargs.copy()
        else:
            return self.kwargs.copy()

    def _iter_data(self):
        """

        :return:
        """
        for sub, plots in self.data.items():
            index = plots['x']
            colors = self._get_colors(num_colors=len(plots['y']))
            kind = self.kind[sub]
            zorder = len(kind) * [None]
            # NOTE: Scatter has zorder=1 by default, so it will be drawn below lines. Setting zorder of scatter to 2
            #  and zorder of lines to 1 to be able to see scatters in plots with dense lines.
            if 'line' in kind and 'scatter' in kind:
                for key, value in enumerate(kind):
                    if value == 'line':
                        zorder[key] = 1
                    elif value == 'scatter':
                        zorder[key] = 2
            for k, (key, values) in enumerate(plots['y'].items()):
                yield sub, k, index[key], key, values, colors[k], kind[k], zorder[k]

    def _make_plot(self):
        """

        :return:
        """
        it = self._iter_data()

        for (subplot, k, x, label, y, colors, kind, zorder) in it:
            ax = self._get_ax(subplot)
            kwargs = self._get_kwargs(kind)

            if kind == 'scatter':
                if is_list_like(kwargs['marker']):
                    kwargs['marker'] = kwargs['marker'][subplot][k]
                if is_list_like(kwargs['s']):
                    kwargs['s'] = kwargs['s'][subplot][k]
            kwargs['color'] = colors
            kwargs['legend'] = False
            kwargs['ax'] = ax
            if zorder is not None:
                kwargs['zorder'] = zorder

            data = {
                0: {
                    'x': {label: x},
                    'y': {label: y}
                }
            }

            plot_obj = MultiPlot.return_plot_class(kind)(data, **kwargs)
            plot_obj.generate()

    @staticmethod
    def return_plot_class(arg):
        """

        :param arg:
        :return:
        """
        return {
            'line': LinePlot,
            'dashed': DashedPlot,
            'dotted': DottedPlot,
            'dashdot': DashDotPlot,
            'step': StepPlot,
            'scatter': ScatterPlot
        }[arg]


class LinePlot(BasePlot):
    """"""
    def __init__(self, data, **kwargs):
        """Constructor method"""
        super().__init__(data, **kwargs)

    def _make_plot(self):
        """

        :return:
        """
        plot = self._plot
        it = self._iter_data()

        for (k, x, label, y, color) in it:
            ax = self._get_ax(k)
            kwargs = self.kwargs.copy()

            x_range = kwargs.pop('x_range', None)

            kwargs['color'] = color
            kwargs['label'] = label

            new_lines = plot(ax, x, y, **kwargs)
            self._add_legend_handle(new_lines[0], label)

            if x_range is None:
                lines = _get_all_lines(ax)
                left, right = _get_x_lim(lines)
            else:
                left, right = x_range
            ax.set_xlim(left, right)


class DashedPlot(LinePlot):
    """"""
    def __init__(self, data, **kwargs):
        """Constructor method"""
        super().__init__(data, **kwargs)

        self.kwargs['linestyle'] = 'dashed'


class DottedPlot(LinePlot):
    """"""
    def __init__(self, data, **kwargs):
        """Constructor method"""
        super().__init__(data, **kwargs)

        self.kwargs['linestyle'] = 'dotted'


class DashDotPlot(LinePlot):
    """"""
    def __init__(self, data, **kwargs):
        """Constructor method"""
        super().__init__(data, **kwargs)

        self.kwargs['linestyle'] = 'dashdot'


class StepPlot(BasePlot):
    """"""
    def __init__(self, data, **kwargs):
        """Constructor method"""
        super().__init__(data, **kwargs)

    def _make_plot(self):
        """

        :return:
        """
        plot = self._plot
        it = self._iter_data()

        for (k, x, label, y, color) in it:
            ax = self._get_ax(k)
            kwargs = self.kwargs.copy()

            x_range = kwargs.pop('x_range', None)

            kwargs['color'] = color
            kwargs['label'] = label

            new_lines = plot(ax, x, y, **kwargs)
            self._add_legend_handle(new_lines[0], label)

            if x_range is None:
                lines = _get_all_lines(ax)
                left, right = _get_x_lim(lines)
            else:
                left, right = x_range
            ax.set_xlim(left, right)

    @classmethod
    def _plot(cls, ax, x, y, style=None, **kwargs):
        """

        :param ax:
        :param x:
        :param y:
        :param style:
        :param kwargs:
        :return:
        """
        if style is not None:
            args = (x, y, style)
        else:
            args = (x, y)
        return ax.step(*args, **kwargs)


class ScatterPlot(BasePlot):
    """"""
    def __init__(self, data, marker='x', **kwargs):
        """Constructor method"""
        super().__init__(data, **kwargs)

        self.marker = marker
        if is_list_like(self.marker) and len(self.marker) != self.n_plots:
            msg = f"The length of 'marker' must equal the number of columns if using 'marker' of type " \
                  f"'list'.\nlength of marker = {len(self.marker)}\nnumber of columns = {self.n_plots}"
            raise ValueError(msg)

    def _make_plot(self):
        """

        :return:
        """
        plot = self._plot
        it = self._iter_data()

        for (k, x, label, y, color) in it:
            ax = self._get_ax(k)
            kwargs = self.kwargs.copy()

            x_range = kwargs.pop('x_range', None)

            if is_list_like(self.marker):
                marker = self.marker[k]
            else:
                marker = self.marker
            kwargs['marker'] = marker
            marker_size = kwargs.get('size')
            if marker_size is None:
                marker_size = kwargs.pop('marker_size', None)
            if marker_size is not None:
                if is_list_like(marker_size):
                    kwargs['s'] = marker_size[k]
                else:
                    kwargs['s'] = marker_size

            # TODO: If we additionally supply the marker_color ('c') keyword this will lead to an error during plotting
            kwargs['color'] = color
            kwargs['label'] = label

            new_lines = plot(ax, x, y, **kwargs)
            self._add_legend_handle(new_lines, label)

            if x_range is None:
                lines = _get_all_lines(ax)
                bboxes = _get_all_bboxes(ax)
                left, right = _get_x_lim(lines, bboxes=bboxes)
            else:
                left, right = x_range
            ax.set_xlim(left, right)

    @classmethod
    def _plot(cls, ax, x, y, style=None, **kwargs):
        """

        :param ax:
        :param x:
        :param y:
        :param style:
        :param kwargs:
        :return:
        """
        if style is not None:
            args = (x, y, style)
        else:
            args = (x, y)
        return ax.scatter(*args, **kwargs)


def _flatten(axes):
    """

    :param axes:
    :return:
    """
    if not is_list_like(axes):
        return np.array([axes])
    elif isinstance(axes, np.ndarray):
        return axes.ravel()
    return np.array(axes)


def _get_all_lines(ax):
    """

    :param ax:
    :return:
    """
    lines = ax.get_lines()

    if hasattr(ax, 'right_ax'):
        lines += ax.right_ax.get_lines()

    if hasattr(ax, 'left_ax'):
        lines += ax.left_ax.get_lines()

    return lines


def _get_all_bboxes(ax):
    """

    :param ax:
    :return:
    """
    bboxes = []
    for collection in ax.collections:
        bboxes.append(collection.axes.dataLim)

    return bboxes


def _get_layout(n_plots, layout=None, layout_type='grid'):
    """

    :param n_plots:
    :param layout:
    :param layout_type:
    :return:
    """
    if layout is not None:
        if not isinstance(layout, (list, tuple)) or len(layout) != 2:
            raise ValueError("Layout must be a tuple of (rows, columns)")

        n_rows, n_cols = layout

        def ceil_(x):
            """

            :param x:
            :return:
            """
            return int(ceil(x))

        if n_rows == -1 and n_cols > 0:
            layout = n_rows, n_cols = (ceil_(float(n_plots) / n_cols), n_cols)
        elif n_cols == -1 and n_rows > 0:
            layout = n_rows, n_cols = (n_rows, ceil_(float(n_plots) / n_rows))
        elif n_cols <= 0 and n_rows <= 0:
            msg = "At least one dimension of layout must be positive"
            raise ValueError(msg)

        if n_rows * n_cols < n_plots:
            raise ValueError(f"Layout of {n_rows}x{n_cols} must be larger than required size {n_plots}")

        return layout

    if layout_type == 'single':
        return 1, 1
    if layout_type == 'horizontal':
        return 1, n_plots
    if layout_type == 'vertical':
        return n_plots, 1

    layouts = {1: (1, 1), 2: (1, 2), 3: (2, 2), 4: (2, 2)}
    try:
        return layouts[n_plots]
    except KeyError:
        k = 1
        while k ** 2 < n_plots:
            k += 1

        if (k - 1) * k >= n_plots:
            return (k - 1), k
        else:
            return k, k


def _get_standard_colors(num_colors=None, colormap=None, color_type='default', color=None):
    """

    :param num_colors:
    :param colormap:
    :param color_type:
    :param color:
    :return:
    """
    import matplotlib.pyplot as plt

    if color is None and colormap is not None:
        if isinstance(colormap, str):
            cmap = colormap
            colormap = cm.get_cmap(colormap)
            if colormap is None:
                raise ValueError(f"Colormap {cmap} is not recognized")
        colors = [colormap[k] for k in range(num_colors)]
    elif color is not None:
        if colormap is not None:
            warnings.warn("'color' and 'colormap' cannot be used simultaneously. Using 'color'...")
        colors = (list(color) if is_list_like(color) and not isinstance(color, dict) else color)
    else:
        if color_type == 'default':
            try:
                colors = [c['color'] for c in list(plt.rcParams['axes.prop_cycle'])]
            except KeyError:
                colors = list(plt.rcParams.get('axes.color_cycle', list('bgrcmyk')))
            if isinstance(colors, str):
                colors = list(colors)

            colors = colors[:num_colors]
        elif color_type == 'random':
            def random_color(column):
                """

                :param column:
                :return:
                """
                rs = random_state(column)
                return rs.rand(3).tolist()

            colors = [random_color(num) for num in range(num_colors)]
        else:
            raise ValueError("'color_type' must be either 'default' or 'random'")

    if isinstance(colors, str):
        conv = matplotlib.colors.ColorConverter()

        def _maybe_valid_colors(colors):
            """

            :param colors:
            :return:
            """
            try:
                [conv.to_rgba(c) for c in colors]
                return True
            except ValueError:
                return False

        maybe_single_color = _maybe_valid_colors([colors])
        maybe_color_cycle = _maybe_valid_colors(list(colors))
        if maybe_single_color and maybe_color_cycle and len(colors) > 1:
            hex_color = [c['color'] for c in list(plt.rcParams['axes.prop_cycle'])]
            colors = [hex_color[int(colors[1])]]
        elif maybe_color_cycle:
            colors = [colors]
        else:
            colors = [colors]

    if len(colors) < num_colors:
        try:
            multiple = num_colors // len(colors) - 1
        except ZeroDivisionError:
            raise ValueError("Invalid color argument: ''")
        mod = num_colors % len(colors)

        colors += multiple * colors
        colors += colors[:mod]

    return colors


def _get_x_lim(lines, bboxes=None):
    """

    :param lines:
    :param bboxes:
    :return:
    """
    left, right = np.inf, -np.inf
    for line in lines:
        x = line.get_xdata(orig=False)
        left = min(np.nanmin(x), left)
        right = max(np.nanmax(x), right)
    if bboxes is not None:
        for bbox in bboxes:
            left = min(bbox.xmin, left)
            right = max(bbox.xmax, right)
    return left, right


def _subplots(n_axes=None, squeeze=True, subplot_kwargs=None, layout=None, layout_type='grid', **kwargs):
    """

    :param n_axes:
    :param squeeze:
    :param subplot_kwargs:
    :param layout:
    :param layout_type:
    :param kwargs:
    :return:
    """
    import matplotlib.pyplot as plt

    if subplot_kwargs is None:
        subplot_kwargs = {}

    if not kwargs:
        x_data = None
    else:
        interactive = kwargs.pop('interactive', False)
        if interactive:
            x_data = kwargs.pop('x_data', None)
            if x_data is None:
                raise KeyError("Argument 'x_data' is missing. If interactive flag is set to True, 'x_data' needs to be"
                               " supplied as well")
        else:
            x_data = None

    fig = plt.figure(**kwargs)
    sliders = []
    if x_data is not None:
        plt.subplots_adjust(left=.05, bottom=.1, right=.95, top=.98, wspace=.2, hspace=.2)

    n_rows, n_cols = _get_layout(n_axes, layout=layout, layout_type=layout_type)
    n_plots = n_rows * n_cols

    ax_arr = np.empty(n_plots, dtype=object)

    ax0 = fig.add_subplot(n_rows, n_cols, 1, **subplot_kwargs)
    if x_data is not None:
        sliders.append(_slider(ax0, x_data[0]))

    ax_arr[0] = ax0

    for k in range(1, n_plots):
        kwds = subplot_kwargs.copy()
        if k >= n_axes:
            kwds['sharex'] = None
            kwds['sharey'] = None
        ax = fig.add_subplot(n_rows, n_cols, k + 1, **kwds)
        if x_data is not None and k < n_axes:
            sliders.append(_slider(ax, x_data[k]))
        ax_arr[k] = ax

    if n_axes != n_plots:
        for ax in ax_arr[n_axes:]:
            ax.set_visible(False)

    if squeeze:
        if n_plots == 1:
            axes = ax_arr[0]
        else:
            axes = ax_arr.reshape(n_rows, n_cols).squeeze()
    else:
        axes = ax_arr.reshape(n_rows, n_cols)

    return fig, axes, sliders


def _slider(ax, data, epsilon=1e-8):
    """

    :param ax:
    :param data:
    :param epsilon:
    :return:
    """
    import matplotlib.pyplot as plt
    from matplotlib.widgets import Slider

    def x_aligned_axes(y_distance, width, **kwargs):
        """

        :param y_distance:
        :param width:
        :param kwargs:
        :return:
        """
        return plt.axes([ax.get_position().x0, ax.get_position().y0 - y_distance, ax.get_position().width, width],
                        **kwargs)

    ax_slider = x_aligned_axes(.05, .02)

    expected = next(iter(data.values()))
    all_equal = all(
        (value == expected).all() if value.size == expected.size else value == expected for value in data.values())
    if all_equal:
        x_data = expected
    else:
        index = np.argmax(value.size for value in data.values())
        x_data = list(data.values())[index]
    n_data = len(x_data)
    diff = np.diff(x_data)
    unique_diff = np.all(np.abs(diff - diff[0]) < epsilon)

    if unique_diff:
        slider = Slider(ax_slider, '', x_data[0], x_data[-1], valinit=x_data[-1], valstep=diff[0])
    else:
        slider = Slider(ax_slider, '', 0, n_data - 1, valinit=n_data - 1, valstep=1)

    return slider


def _update_slider(fig, ax, slider, data):
    """

    :param fig:
    :param ax:
    :param slider:
    :param data:
    :return:
    """
    x_data = data['x']
    y_data = data['y']

    keys = y_data.keys()
    lines = ax.get_lines()

    def f(val):
        """

        :param val:
        :return:
        """
        args = []
        for key in keys:
            mask = x_data[key] > val
            arg = y_data[key].copy()
            arg[mask] = np.nan
            args.append(arg)
        return args

    def update(val):
        """

        :param val:
        :return:
        """
        args = f(val)
        for k, arg in enumerate(args):
            lines[k].set_ydata(arg)
        fig.canvas.draw_idle()

    slider.on_changed(update)


__all__ = [
    'MultiPlot',
    'LinePlot',
    'StepPlot',
    'DashedPlot',
    'DottedPlot',
    'DashDotPlot',
    'ScatterPlot'
]
